from __future__ import annotations

from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Qt, Signal


class QIntOrNoneValidator(QtGui.QIntValidator):
    def validate(self, input: str, pos: int) -> tuple[QtGui.QValidator.State, str, int]:
        if input == "":
            return QtGui.QValidator.State.Acceptable, input, pos
        return super().validate(input, pos)


class QIntOrNoneEdit(QtW.QLineEdit):
    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self.setValidator(QIntOrNoneValidator())
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.textChanged.connect(self.ensureVisible)

    def value(self):
        return int(self.text()) if self.text() else None

    def setValue(self, val: int | None):
        self.setText(str(val) if val is not None else "")

    def increment(self):
        val = self.value()
        if val is not None:
            self.setValue(val + 1)

    def decrement(self):
        val = self.value()
        if val is not None:
            self.setValue(max(val - 1, 0))

    def event(self, event: QtCore.QEvent) -> bool:
        if event.type() == QtCore.QEvent.Type.KeyPress:
            event = QtGui.QKeyEvent(event)
            key = event.key()
            mod = event.modifiers()
            if mod != Qt.KeyboardModifier.NoModifier:
                return super().event(event)
            if key == Qt.Key.Key_Up:
                self.increment()
                return True
            elif key == Qt.Key.Key_Down:
                self.decrement()
                return True
        return super().event(event)

    def wheelEvent(self, a0: QtGui.QWheelEvent) -> None:
        if a0.angleDelta().y() > 0:
            self.increment()
        else:
            self.decrement()
        return super().wheelEvent(a0)

    def ensureVisible(self, text: str) -> QtCore.QSize:
        metrics = QtGui.QFontMetrics(self.font())
        w = metrics.width("9") * len(text) + 6
        self.setFixedWidth(w)


class QSelectionRangeEdit(QtW.QWidget):
    sliceChanged = Signal(object)

    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self.setLayout(QtW.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        inner = QtW.QWidget()
        self.layout().addWidget(inner)
        layout = QtW.QHBoxLayout()
        layout.setContentsMargins(8, 0, 8, 0)
        layout.setSpacing(0)
        inner.setLayout(layout)
        self._r_start = QIntOrNoneEdit()
        self._r_stop = QIntOrNoneEdit()
        self._c_start = QIntOrNoneEdit()
        self._c_stop = QIntOrNoneEdit()
        self._r_start.setToolTip("Row starting index")
        self._r_stop.setToolTip("Row stopping index")
        self._c_start.setToolTip("Column starting index")
        self._c_stop.setToolTip("Column stopping index")
        self._r_start.textChanged.connect(self._rstart_changed)
        self._r_colon = _label(":")
        self._r_stop.textChanged.connect(self._rstop_changed)
        self._c_start.textChanged.connect(self._cstart_changed)
        self._c_colon = _label(":")
        self._c_stop.textChanged.connect(self._cstop_changed)
        layout.addWidget(self._r_start)
        layout.addWidget(self._r_colon)
        layout.addWidget(self._r_stop)
        layout.addWidget(_label(", "))
        layout.addWidget(self._c_start)
        layout.addWidget(self._c_colon)
        layout.addWidget(self._c_stop)

    def slice(self) -> tuple[slice, slice]:
        rsl = slice(self._r_start.value(), self._r_stop.value())
        csl = slice(self._c_start.value(), self._c_stop.value())
        return rsl, csl

    def setSlice(self, sl: tuple[slice, slice]):
        rsl, csl = sl
        rstart, rstop = rsl.start, rsl.stop
        cstart, cstop = csl.start, csl.stop
        self._r_start.setValue(rstart)
        self._r_stop.setValue(rstop)
        self._c_start.setValue(cstart)
        self._c_stop.setValue(cstop)
        if rstart is not None and rstop is not None and rstop == rstart + 1:
            self._r_stop.hide()
            self._r_colon.hide()
        else:
            self._r_stop.show()
            self._r_colon.show()
        if cstart is not None and cstop is not None and cstop == cstart + 1:
            self._c_stop.hide()
            self._c_colon.hide()
        else:
            self._c_stop.show()
            self._c_colon.show()
        return None

    def _rstart_changed(self, txt: str):
        rstop = self._r_stop.text()
        if txt and rstop:
            if not self._r_stop.isVisible() or int(rstop) <= int(txt):
                self._r_stop.setText(str(int(txt) + 1))
        return self.sliceChanged.emit(self.slice())

    def _rstop_changed(self, txt: str):
        rstart = self._r_start.text()
        if txt and rstart and int(rstart) >= int(txt):
            int_rstop = int(txt)
            if int_rstop > 1:
                self._r_start.setText(str(int_rstop - 1))
            else:
                self._r_start.setText("0")
                self._r_stop.setText("1")
        return self.sliceChanged.emit(self.slice())

    def _cstart_changed(self, txt: str):
        cstop = self._c_stop.text()
        if txt and cstop:
            if not self._c_stop.isVisible() or int(cstop) <= int(txt):
                self._c_stop.setText(str(int(txt) + 1))
        return self.sliceChanged.emit(self.slice())

    def _cstop_changed(self, txt: str):
        cstart = self._c_start.text()
        if txt and cstart and int(cstart) >= int(txt):
            int_cstop = int(txt)
            if int_cstop > 1:
                self._c_start.setText(str(int_cstop - 1))
            else:
                self._c_start.setText("0")
                self._c_stop.setText("1")
        return self.sliceChanged.emit(self.slice())


def _label(text: str):
    label = QtW.QLabel(text)
    label.setAlignment(Qt.AlignmentFlag.AlignCenter)
    return label



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tabulous.widgets import TableViewerBase


def show_keymap(viewer: TableViewerBase):
    """Show key map widget"""
    viewer._qwidget.showKeyMap()


def close_window(viewer: TableViewerBase):
    """Close window"""
    viewer._qwidget.close()


def new_window(viewer: TableViewerBase):
    """Create a new window"""
    new = viewer.__class__()
    pos = viewer._qwidget.pos()
    pos.setX(pos.x() + 20)
    pos.setY(pos.y() + 20)
    new._qwidget.move(pos)
    return new._qwidget.activateWindow()


def toggle_toolbar(viewer: TableViewerBase):
    """Toggle toolbar visibility"""
    viewer._qwidget.toggleToolBarVisibility()


def toggle_console(viewer: TableViewerBase):
    """Toggle QtConsole visibility"""
    return viewer._qwidget.toggleConsoleVisibility()


def toggle_fullscreen(viewer: TableViewerBase):
    """Toggle fullscreen"""
    viewer._qwidget.toggleWindowState()


def show_command_palette(viewer: TableViewerBase):
    """Show the command palette."""
    viewer._qwidget.showCommandPalette()


def focus_table(viewer: TableViewerBase):
    """Move focus to the table."""
    viewer.native.setCellFocus()


def toggle_focus(viewer: TableViewerBase):
    """Toggle focus between table and command palette."""
    table = viewer.current_table
    qviewer = viewer._qwidget
    if table is None:
        return
    if table._qwidget._qtable_view.hasFocus():
        console = qviewer._console_widget
        if console is not None and console.isActive():
            console.setFocus()
    else:
        qviewer.setCellFocus()
    return


def add_text_edit(viewer: TableViewerBase):
    """Add a note (text edit) dock widget."""
    from tabulous import TableViewer, TableViewerWidget
    from tabulous._qt._qt_const import MonospaceFontFamily
    from qtpy import QtWidgets as QtW, QtGui

    txt_edit = QtW.QTextEdit()
    txt_edit.setWordWrapMode(QtGui.QTextOption.WrapMode.NoWrap)
    txt_edit.setFont(QtGui.QFont(MonospaceFontFamily, 10))

    metrics = txt_edit.fontMetrics()
    txt_edit.setTabStopWidth(4 * metrics.width(" "))

    if isinstance(viewer, TableViewer):
        viewer.add_dock_widget(txt_edit, name="Note")
    elif isinstance(viewer, TableViewerWidget):
        viewer.add_widget(txt_edit, name="Note")
    else:
        raise TypeError(f"Cannot add widget to {type(viewer)}")
    return None


def show_preference(viewer: TableViewerBase):
    """Show preference dialog"""
    viewer._qwidget.showPreferenceDialog()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable
import logging
from qt_command_palette import get_palette
from tabulous import commands as cmds
from tabulous._utils import get_config
from ._mainwidgets import QMainWindow, QMainWidget

if TYPE_CHECKING:
    from ._base import _QtMainWidgetBase
    from tabulous.widgets import TableViewerBase

logger = logging.getLogger("tabulous")


def _command_to_viewer_function(
    f: Callable[[TableViewerBase], Any]
) -> Callable[[_QtMainWidgetBase], Any]:
    def wrapper(self: _QtMainWidgetBase):
        logger.debug(f"Command: {f.__module__.split('.')[-1]}.{f.__name__}")
        return f(self._table_viewer)

    wrapper.__doc__ = f.__doc__
    return wrapper


def load_all_commands():

    palette = get_palette("tabulous")

    window_group = palette.add_group("Window")
    file_group = palette.add_group("File")
    table_group = palette.add_group("Table")
    tab_group = palette.add_group("Tab")
    analysis_group = palette.add_group("Analysis")
    view_group = palette.add_group("View")
    plot_group = palette.add_group("Plot")
    selection_group = palette.add_group("Selection")
    column_group = palette.add_group("Column")

    _groups = {
        "window": window_group,
        "file": file_group,
        "table": table_group,
        "tab": tab_group,
        "analysis": analysis_group,
        "view": view_group,
        "plot": plot_group,
        "selection": selection_group,
        "column": column_group,
    }

    kb = get_config().keybindings.copy()

    for mod, cmd in cmds.iter_commands():
        group = _groups[mod]
        group.register(_command_to_viewer_function(cmd), desc=cmd.__doc__)
        if seq := kb.pop(f"{mod}.{cmd.__name__}", None):
            # register to main widgets
            f = _command_to_viewer_function(cmd)
            if isinstance(seq, str):
                QMainWidget._keymap.bind(seq)(f)
                QMainWindow._keymap.bind(seq)(f)
            elif isinstance(seq, list):
                for s in seq:
                    QMainWidget._keymap.bind(s)(f)
                    QMainWindow._keymap.bind(s)(f)

    if kb:
        import warnings

        keys = ", ".join(kb.keys())
        warnings.warn(f"Unrecognized commands: {keys}")


load_all_commands()



from functools import partial
from typing import List, Union, TYPE_CHECKING, Tuple

import logging
import ast
import numpy as np

# NOTE: Axes should be imported here!
from tabulous.widgets import TableBase, TableViewerBase
from tabulous.types import TableData
from tabulous.exceptions import SelectionRangeError
from tabulous._selection_op import SelectionOperator
from tabulous._magicgui import dialog_factory, dialog_factory_mpl, Axes, ToggleSwitch
from typing_extensions import Annotated

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)
_bool = Annotated[bool, {"widget_type": ToggleSwitch}]


@dialog_factory
def summarize_table(
    table: TableBase,
    methods: List[str],
    new_table: _bool = False,
    at_selection: _bool = False,
):
    if len(methods) == 0:
        raise ValueError("Select at least one method.")
    df = table.data
    if at_selection:
        sels = table.selections
        if len(sels) != 1:
            raise SelectionRangeError("Only single selection is allowed.")
        sel = sels[0]
        df = df.iloc[sel]
    return df.agg(methods), new_table


@dialog_factory
def cut(df: TableData, column: str, bins: str, close_right: _bool = True) -> TableData:
    import pandas as pd

    bins = ast.literal_eval(bins)
    ds = pd.cut(df[column], bins=bins, right=close_right)
    return ds


@dialog_factory
def concat(
    viewer: TableViewerBase,
    names: List[str],
    axis: int,
    ignore_index: _bool = False,
) -> TableData:
    import pandas as pd

    if len(names) < 2:
        raise ValueError("At least two tables are required.")
    dfs = [viewer.tables[name].data for name in names]
    out: pd.DataFrame = pd.concat(dfs, axis=axis, ignore_index=ignore_index)
    if out.index.duplicated().any():
        out.index = range(out.index.size)
    if out.columns.duplicated().any():
        from tabulous._pd_index import char_range_index

        out.columns = char_range_index(out.columns.size)
    return out


@dialog_factory
def merge(
    merge: TableBase,
    with_: TableBase,
    how: str = "inner",
    on: List[str] = (),
) -> TableData:
    if merge is with_:
        raise ValueError("Cannot merge a table with itself.")
    df0 = merge.data
    df1 = with_.data
    if len(on) == 0:
        on = None
    return df0.merge(df1, how=how, on=on)


@dialog_factory
def fillna(df: TableData, method: str, value) -> TableData:
    if method != "value":
        value = None
    else:
        method = None
    return df.fillna(value=value, method=method)


@dialog_factory
def dropna(df: TableData, axis: int, how: str) -> TableData:
    return df.dropna(axis=axis, how=how, inplace=False)


@dialog_factory
def pivot(df: TableData, index: str, columns: str, values: str) -> TableData:
    return df.pivot(index=index, columns=columns, values=values)


@dialog_factory
def sort(df: TableData, by: List[str], ascending: _bool = True) -> TableData:
    return df.sort_values(by=by, ascending=ascending)


@dialog_factory_mpl
def plot(
    ax: Axes,
    x: SelectionOperator,
    y: SelectionOperator,
    table: TableBase,
    alpha: float = 1.0,
    retain_reference: _bool = False,
):
    from ._plot_models import PlotModel

    model = PlotModel(ax, x, y, table=table, alpha=alpha, ref=retain_reference)
    model.add_data()
    table.plt.draw()
    return True


@dialog_factory_mpl
def bar(
    ax: Axes,
    x: SelectionOperator,
    y: SelectionOperator,
    table: TableBase,
    alpha: float = 1.0,
    retain_reference: _bool = False,
):
    from ._plot_models import BarModel

    model = BarModel(ax, x, y, table=table, alpha=alpha, ref=retain_reference)
    model.add_data()
    table.plt.draw()
    return True


@dialog_factory_mpl
def scatter(
    ax: Axes,
    x: SelectionOperator,
    y: SelectionOperator,
    label: SelectionOperator,
    table: TableBase,
    alpha: float = 1.0,
    retain_reference: _bool = False,
):
    from ._plot_models import ScatterModel

    model = ScatterModel(
        ax, x, y, table=table, label_selection=label, alpha=alpha, ref=retain_reference
    )
    model.add_data()
    table.plt.draw()
    return True


@dialog_factory_mpl
def errorbar(
    ax: Axes,
    x: SelectionOperator,
    y: SelectionOperator,
    xerr: SelectionOperator,
    yerr: SelectionOperator,
    label: SelectionOperator,
    table: TableBase,
    alpha: float = 1.0,
):
    data = table.data
    ydata = _operate_column(y, data)

    xerrdata = _operate_column(xerr, data, default=None)
    yerrdata = _operate_column(yerr, data, default=None)
    if xerrdata is None and yerrdata is None:
        raise ValueError("Either x-error or y-error must be set.")

    labeldata = _operate_column(label, data, default=None)
    if x is None:
        import pandas as pd

        xdata = pd.Series(np.arange(len(ydata)), name="X")
    else:
        xdata = _operate_column(x, data)

    _errorbar = partial(ax.errorbar, alpha=alpha, fmt="o", picker=True)
    if labeldata is None:
        _errorbar(xdata, ydata, xerr=xerrdata, yerr=yerrdata, label=y)
    else:
        unique = labeldata.unique()
        for label_ in unique:
            spec = labeldata == label_
            if xerrdata is not None:
                xerr_ = xerrdata[spec]
            else:
                xerr_ = None
            if yerrdata is not None:
                yerr_ = yerrdata[spec]
            else:
                yerr_ = None
            _errorbar(
                xdata[spec],
                ydata[spec],
                xerr=xerr_,
                yerr=yerr_,
                label=label_,
            )

    table.plt.draw()
    return True


@dialog_factory_mpl
def fill_between(
    ax: Axes,
    x: SelectionOperator,
    y0: SelectionOperator,
    y1: SelectionOperator,
    table: TableBase,
    alpha: float = 1.0,
    retain_reference: _bool = False,
):
    from ._plot_models import FillBetweenModel

    model = FillBetweenModel(
        ax, x, y0, y1, table=table, alpha=alpha, ref=retain_reference
    )
    model.add_data()
    table.plt.draw()
    return True


@dialog_factory_mpl
def fill_betweenx(
    ax: Axes,
    y: SelectionOperator,
    x0: SelectionOperator,
    x1: SelectionOperator,
    table: TableBase,
    alpha: float = 1.0,
    retain_reference: _bool = False,
):
    from ._plot_models import FillBetweenXModel

    model = FillBetweenXModel(
        ax, y, x0, x1, table=table, alpha=alpha, ref=retain_reference
    )
    model.add_data()
    table.plt.draw()
    return True


@dialog_factory_mpl
def hist(
    ax: Axes,
    y: SelectionOperator,
    label: SelectionOperator,
    table: TableBase,
    bins: int = 10,
    range: Tuple[str, str] = ("", ""),
    alpha: float = 1.0,
    density: _bool = False,
    histtype: str = "bar",
):
    from ._plot_models import HistModel

    r0, r1 = range
    if r0 or r1:
        _range = float(r0), float(r1)
    else:
        _range = None
    model = HistModel(
        ax,
        y,
        bins=bins,
        table=table,
        range=_range,
        label_selection=label,
        alpha=alpha,
        density=density,
        histtype=histtype,
    )
    model.add_data()
    ax.axhline(0, color="gray", lw=0.5, alpha=0.5, zorder=-1)
    table.plt.draw()
    return True


@dialog_factory_mpl
def swarmplot(
    ax: Axes,
    x: str,
    y: str,
    table: TableBase,
    csel,
    hue: str = None,
    dodge: _bool = False,
    alpha: Annotated[float, {"min": 0.0, "max": 1.0}] = 1.0,
):
    import seaborn as sns

    data = table.data[csel]
    sns.swarmplot(
        x=x, y=y, data=data, hue=hue, dodge=dodge, alpha=alpha, ax=ax, picker=True
    )
    table.plt.draw()
    return True


@dialog_factory_mpl
def barplot(
    ax: Axes,
    x: str,
    y: str,
    table: TableBase,
    csel,
    hue: str = None,
    dodge: _bool = False,
    alpha: Annotated[float, {"min": 0.0, "max": 1.0}] = 1.0,
):
    import seaborn as sns

    data = table.data[csel]
    sns.barplot(
        x=x, y=y, data=data, hue=hue, dodge=dodge, alpha=alpha, ax=ax, picker=True
    )
    ax.axhline(0, color="gray", lw=0.5, alpha=0.5, zorder=-1)
    table.plt.draw()
    return True


@dialog_factory_mpl
def boxplot(
    ax: Axes,
    x: str,
    y: str,
    table: TableBase,
    csel,
    hue: str = None,
    dodge: _bool = False,
):
    import seaborn as sns

    data = table.data[csel]
    sns.boxplot(x=x, y=y, data=data, hue=hue, dodge=dodge, ax=ax)
    table.plt.draw()
    return True


@dialog_factory_mpl
def boxenplot(
    ax: Axes,
    x: str,
    y: str,
    table: TableBase,
    csel,
    hue: str = None,
    dodge: _bool = False,
):
    import seaborn as sns

    data = table.data[csel]
    sns.boxenplot(x=x, y=y, data=data, hue=hue, dodge=dodge, ax=ax, picker=True)
    table.plt.draw()
    return True


@dialog_factory_mpl
def violinplot(
    ax: Axes,
    x: str,
    y: str,
    table: TableBase,
    csel,
    hue: str = None,
    dodge: _bool = False,
):
    import seaborn as sns

    data = table.data[csel]
    sns.violinplot(x=x, y=y, data=data, hue=hue, dodge=dodge, ax=ax, picker=True)
    table.plt.draw()
    return True


@dialog_factory_mpl
def stripplot(
    ax: Axes,
    x: str,
    y: str,
    table: TableBase,
    csel,
    hue: str = None,
    dodge: _bool = False,
):
    import seaborn as sns

    data = table.data[csel]
    sns.stripplot(x=x, y=y, data=data, hue=hue, dodge=dodge, ax=ax, picker=True)
    table.plt.draw()
    return True


@dialog_factory
def choose_one(choice: str):
    return choice


@dialog_factory
def choose_multiple(choices: List):
    return choices


@dialog_factory
def get_value(x):
    return x


@dialog_factory
def spinbox(min: str = "0", max: str = "1000", step: str = "") -> dict:
    min = int(min)
    max = int(max)
    step = int(step) if step != "" else None
    return dict(min=min, max=max, step=step)


@dialog_factory
def float_spinbox(min: str = "0.0", max: str = "1000.0", step: str = "") -> dict:
    min = float(min)
    max = float(max)
    step = float(step) if step != "" else None
    return dict(min=min, max=max, step=step)


@dialog_factory
def slider(min: str = "0", max: str = "1000", step: str = "") -> dict:
    min = int(min)
    max = int(max)
    step = int(step) if step != "" else None
    return dict(min=min, max=max, step=step)


@dialog_factory
def float_slider(min: str = "0.0", max: str = "1000.0", step: str = "") -> dict:
    min = float(min)
    max = float(max)
    step = float(step) if step != "" else None
    return dict(min=min, max=max, step=step)


@dialog_factory
def checkbox(text: str, checked: _bool = True) -> dict:
    return dict(text=text, value=checked)


@dialog_factory
def radio_button(text: str, checked: _bool = True) -> dict:
    return dict(text=text, value=checked)


__void = object()


def _operate_column(
    op: Union[SelectionOperator, None],
    data: "pd.DataFrame",
    default=__void,
) -> Union["pd.Series", None]:
    if op is None:
        if default is __void:
            raise ValueError("Wrong selection.")
        return default
    import pandas as pd

    ds = op.operate(data)
    if isinstance(ds, pd.DataFrame):
        if len(ds.columns) != 1:
            raise ValueError("Operation must return a single column.")
        ds = ds.iloc[:, 0]
    return ds



from __future__ import annotations
from pathlib import Path
from qtpy import QtWidgets as QtW, QtGui

from tabulous.color import normalize_color
from tabulous._qt._svg import QColoredSVGIcon


ICON_DIR = Path(__file__).parent.parent / "_icons"


class QColoredToolButton(QtW.QToolButton):
    """Tool button with colored icon."""

    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setStyleSheet("QToolButton::menu-indicator { image: none; }")

    def icon(self) -> QColoredSVGIcon:
        """Get colored svg icon."""
        return self._icon

    def setIcon(self, icon: QColoredSVGIcon | str | Path) -> None:
        """Set colored svg icon."""
        if isinstance(icon, (str, Path)):
            icon = QColoredSVGIcon.fromfile(icon)
        self._icon = icon
        return super().setIcon(icon)

    def updateColor(self, color: str):
        """Update svg color."""
        icon = self.icon().colored(color)
        if not self.isEnabled():
            icon = icon.with_converted(self._mix_with_background)
        self.setIcon(icon)
        return None

    def setEnabled(self, a0: bool) -> None:
        if self.isEnabled() != a0:
            # blend icon color with background color
            if a0:
                self.setIcon(self.icon().with_converted(lambda x: x))
            else:
                self.setIcon(self.icon().with_converted(self._mix_with_background))
        return super().setEnabled(a0)

    def backgroundColor(self) -> QtGui.QColor:
        """Get background color."""
        return self.palette().color(self.backgroundRole())

    def _mix_with_background(self, color: QtGui.QColor) -> QtGui.QColor:
        """Mix color with background color."""
        bg = self.palette().color(self.backgroundRole()).getRgb()
        return QtGui.QColor(*normalize_color(color.getRgb()).mix(bg, 0.7))


class QMoreToolButton(QColoredToolButton):
    """Tool button that shows the menu when clicked."""

    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setIcon(ICON_DIR / "more.svg")
        self.setPopupMode(QtW.QToolButton.ToolButtonPopupMode.InstantPopup)
        self.setToolTip("More ...")



from __future__ import annotations
from typing import Literal, TYPE_CHECKING, cast
import weakref
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Qt, Signal

from tabulous._qt._table_stack._start import QStartupWidget
from tabulous._qt._table_stack._utils import create_temporal_line_edit

from tabulous._qt._table._base._table_group import QTableGroup
from tabulous._qt._clickable_label import QClickableLabel
from tabulous._qt._action_registry import QActionRegistry

from tabulous._utils import get_config

if TYPE_CHECKING:
    from tabulous._qt._table import QBaseTable, QMutableTable
    from tabulous._qt._mainwindow._base import _QtMainWidgetBase
    from tabulous._qt._table_stack._overlay import QOverlayWidget


class QTabbedTableStack(QtW.QTabWidget, QActionRegistry[int]):
    """Tab widget used for table stack."""

    currentTableChanged = Signal(int)  # index
    itemDropped = Signal(object)  # dropped item info
    itemMoved = Signal(int, int)  # source index, destination index
    tableRenamed = Signal(int, str)  # index
    tableRemoved = Signal(int)  # index
    tablePassed = Signal(object, int, object)  # source widget, tab_id, target widget
    resizedSignal = Signal()

    def __init__(
        self,
        parent=None,
        tab_position: str = "top",
    ):
        QtW.QTabWidget.__init__(self, parent)
        QActionRegistry.__init__(self)

        self.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )

        from . import _tabbar

        if tab_position == "top":
            self.setTabBar(_tabbar.QTabulousTabBar(self))
        elif tab_position == "left":
            self.setTabBar(_tabbar.QLeftSideBar(self))
            self.setTabPosition(QtW.QTabWidget.TabPosition.West)
        elif tab_position == "bottom":
            self.setTabBar(_tabbar.QTabulousTabBar(self))
            self.setTabPosition(QtW.QTabWidget.TabPosition.South)
        elif tab_position == "right":
            self.setTabBar(_tabbar.QRightSideBar(self))
            self.setTabPosition(QtW.QTabWidget.TabPosition.East)
        else:
            raise ValueError(f"Unknown position {tab_position!r}.")

        self.setAcceptDrops(True)
        self.setMovable(True)

        self.tabBar().customContextMenuRequested.connect(self.showContextMenu)
        self.currentChanged.connect(self.currentTableChanged.emit)
        self.tabCloseRequested.connect(self.tableRemoved.emit)
        # NOTE: arguments are not (from, to). Bug in Qt??
        self.tabBar().tabMoved.connect(lambda a, b: self.itemMoved.emit(b, a))
        self.tabBarDoubleClicked.connect(self.enterEditingMode)

        # NOTE: this is needed to correctly refocus table groups.
        self.tabBarClicked.connect(self.setCurrentIndex)

        # add overlay widget
        from ._overlay import QOverlayWidget

        self._overlay = QOverlayWidget(self)
        self._notifier = QOverlayWidget(self, duration=200)
        self._notifier.setAnchor("top_right")

        # temporal QLineEdit for editing tabs
        self._line: QtW.QLineEdit | None = None

        self._entering_tab_index: int | None = None

        # "new tab" button
        tb = QtW.QToolButton()
        tb.setText("+")
        tb.setFont(QtGui.QFont("Arial", 12, weight=15))
        tb.setToolTip("New spreadsheet")
        tb.clicked.connect(lambda: self.parent()._table_viewer.add_spreadsheet())
        self.setCornerWidget(tb)
        self.addEmptyWidget()

    # fmt: off
    if TYPE_CHECKING:
        def parent(self) -> _QtMainWidgetBase: ...
        def parentWidget(self) -> _QtMainWidgetBase: ...
    # fmt: on

    def addEmptyWidget(self):
        """Add empty widget to stack."""
        assert self.count() == 0
        startup = QStartupWidget()
        self.addTab(startup, "")
        self.tabBar().hide()
        if parent := self.parentWidget():
            if toolbar := parent._toolbar:
                toolbar._corner_widget.hide()
        return

    def isEmpty(self) -> bool:
        return self.count() == 1 and isinstance(self.widget(0), QStartupWidget)

    def addTable(self, table: QBaseTable, name: str = "None"):
        """Add `table` to stack as name `name`."""
        if self.isEmpty():
            self.removeTab(0)
            self.tabBar().show()
            if toolbar := self.parentWidget()._toolbar:
                toolbar._corner_widget.setVisible(get_config().window.selection_editor)
        self.addTab(table, name)
        table._qtable_view.resizedSignal.connect(self.resizedSignal.emit)
        table.selectionChangedSignal.connect(
            lambda: self.updateSelectionEdit(table.selections())
        )
        return None

    def takeTable(self, index: int) -> QBaseTable:
        """Remove table at `index` and return it."""
        table = self.tableAtIndex(index)
        self.untileTable(index)
        self.removeTab(index)
        if self.count() == 0:
            self.addEmptyWidget()
        else:
            self.parentWidget().setCellFocus()
        return table

    def renameTable(self, index: int, name: str):
        """Rename table at `index` to `name`."""
        return self.setTabText(index, name)

    def tableIndex(self, table: QBaseTable) -> int:
        """Get the index of `table`."""
        for i in range(self.count()):
            data = self.tableAtIndex(i)
            if data is table:
                break
        else:
            raise ValueError(f"Table {table!r} not found.")
        return i

    def tableAtIndex(self, i: int) -> QBaseTable:
        """Get the table at `i`."""
        wdt = self.widget(i)
        if isinstance(wdt, QStartupWidget):
            return None
        if isinstance(wdt, QTableGroup):
            wdt = cast(QTableGroup, wdt)
            idx = self._tab_index_to_group_index(i)
            return wdt.tables[idx]
        return wdt

    def tableAt(self, pos: QtCore.QPoint) -> QBaseTable | None:
        """Return table at position."""
        index = self.tabBar().tabAt(pos)
        # TODO: bug when position is east or south
        if index == -1:
            return None
        return self.tableAtIndex(index)

    def moveTable(self, src: int, dst: int):
        """Move table from `src` to `dst`."""
        return self.tabBar().moveTab(src, dst)

    def updateSelectionEdit(self, sel: list[tuple[slice, slice]]):
        if sel:
            if toolbar := self.parentWidget()._toolbar:
                corner = toolbar._corner_widget
                corner.blockSignals(True)
                corner.setSlice(sel[-1])
                corner.blockSignals(False)

    def dragEnterEvent(self, e: QtGui.QDragEnterEvent) -> None:
        # This override is necessary for accepting drops from files.
        e.accept()
        source = e.source()
        if source is None:
            return
        if source.parent() is not self:
            return

        self._entering_tab_index = self.indexOf(self.widget(self._moving_tab_index))
        return None

    def dropEvent(self, e: QtGui.QDropEvent) -> None:
        mime = e.mimeData()
        text = mime.text()
        if text:
            # File is dropped.
            for path in text.splitlines():
                if path:
                    # properly format path
                    if path.startswith("file:///"):
                        # NOTE: this is needed! lstrip-only does not work
                        path = path.lstrip("file:///")
                    elif path.startswith("file:"):
                        path = path.lstrip("file:")
                    self.itemDropped.emit(path)

        source = e.source()
        if source is None:
            return
        source_widget: QTabbedTableStack = source.parent()
        if not isinstance(source_widget, QTabbedTableStack):
            return
        tab_id = source_widget._entering_tab_index
        if source_widget is self:
            return super().dropEvent(e)
        if tab_id is None:
            return None
        # Tab from other stack is dropped.
        e.setDropAction(Qt.DropAction.MoveAction)
        e.accept()

        self.tablePassed.emit(source_widget, tab_id, self)
        return super().dropEvent(e)

    def mouseMoveEvent(self, e: QtGui.QMouseEvent):
        if self.isEmpty():
            return
        self._startDrag(e.pos())

    def _startDrag(self, pos: QtCore.QPoint):
        pos_global = self.mapToGlobal(pos)
        tabbar = self.tabBar()
        pos_intab = tabbar.mapFromGlobal(pos_global)
        self._moving_tab_index = tabbar.tabAt(pos)
        if self._moving_tab_index < 0:
            return

        tabrect = self.tabRect(self._moving_tab_index)

        pixmap = QtGui.QPixmap(tabrect.size())
        tabbar.render(pixmap, QtCore.QPoint(), QtGui.QRegion(tabrect))
        mime = QtCore.QMimeData()
        drag = QtGui.QDrag(tabbar)
        drag.setMimeData(mime)
        drag.setPixmap(pixmap)
        cursor = QtGui.QCursor(Qt.CursorShape.OpenHandCursor)
        drag.setHotSpot(pos - pos_intab)
        drag.setDragCursor(cursor.pixmap(), Qt.DropAction.MoveAction)
        drag.exec_(Qt.DropAction.MoveAction)

    def mousePressEvent(self, e: QtGui.QMouseEvent) -> None:
        if self._line is not None:
            self._line.setHidden(True)
        return super().mousePressEvent(e)

    def mouseDoubleClickEvent(self, e: QtGui.QMouseEvent) -> None:
        if e.button() != Qt.MouseButton.LeftButton:
            return
        return super().mouseDoubleClickEvent(e)

    def dragLeaveEvent(self, e: QtGui.QDragLeaveEvent):
        e.accept()
        return None

    def resizeEvent(self, e: QtGui.QResizeEvent) -> None:
        self.resizedSignal.emit()
        return super().resizeEvent(e)

    def tabRect(self, index: int):
        """Get QRect of the tab at index."""
        rect = self.tabBar().tabRect(index)

        # NOTE: East/South tab returns wrong value (Bug in Qt?)
        if self.tabPosition() == QtW.QTabWidget.TabPosition.East:
            w = self.rect().width() - rect.width()
            rect.translate(w, 0)
        elif self.tabPosition() == QtW.QTabWidget.TabPosition.South:
            h = self.rect().height() - rect.height()
            rect.translate(0, h)

        return rect

    def enterEditingMode(self, index: int):
        """Enter edit table name mode."""
        if self.isEmpty():
            return
        rect = self.tabRect(index)

        # set geometry
        line = create_temporal_line_edit(rect, self, self.tabText(index))
        self._line = line

        @self._line.editingFinished.connect
        def _():
            self._line.setHidden(True)
            text = self._line.text()
            if text:
                # do not rename if text is empty
                self.setTabText(index, text)
                self.tableRenamed.emit(index, text)
            self.parentWidget().setCellFocus()

        return None

    def showContextMenu(self, pos: QtCore.QPoint) -> None:
        """Execute contextmenu."""
        table = self.tableAt(pos)
        if table is None:
            return
        index = self.tableIndex(table)
        return self.execContextMenu(self.mapToGlobal(pos), index)

    def notifyEditability(self):
        """Show a notification saying that the table is not editable."""
        idx = self.currentIndex()
        qtable = self.tableAtIndex(idx)
        name = self.tabText(idx)
        return self.notifyByWidget(QEditabilityNotifier(qtable, name))

    def notifyLatestVersion(self, version: str):
        return self.notifyByWidget(QLatestVersionNotifier(version))

    def notifyNotSaved(self, idx: int):
        name = self.tabText(idx)
        return self.notifyByWidget(QNotSavedNotifier(name))

    def notifyByWidget(self, widget: QtW.QWidget):
        self._notifier.addWidget(widget)
        self._notifier.show()
        self._notifier.hideLater()
        return None

    def setCurrentIndex(self, index: int):
        """Set current active index."""
        wdt = self.widget(index)
        if isinstance(wdt, QTableGroup):
            wdt = cast(QTableGroup, wdt)
            wdt.setFocusedIndex(self._tab_index_to_group_index(index))
        table = self.tableAtIndex(index)
        self.updateSelectionEdit(table.selections())
        return super().setCurrentIndex(index)

    def tileTables(
        self,
        indices: list[int],
        orientation: Literal["horizontal", "vertical"] = "horizontal",
    ):
        """Merge tables at indices."""
        # strict check of indices
        if len(indices) < 2:
            raise ValueError("Need at least two tables to merge.")
        elif not all(0 <= idx < self.count() for idx in indices):
            raise IndexError("Table indices out of range.")
        elif not all(isinstance(idx, int) for idx in indices):
            raise TypeError("Indices must be integers.")
        elif len(set(indices)) != len(indices):
            raise ValueError("Duplicate indices.")

        # check orientation
        if orientation == "horizontal":
            ori = Qt.Orientation.Horizontal
        elif orientation == "vertical":
            ori = Qt.Orientation.Vertical
        else:
            raise ValueError("Orientation must be 'horizontal' or 'vertical'.")

        current_index = self.currentIndex()
        indices = sorted(indices)

        # untile tables before tiling with others
        tables: list[QBaseTable] = []
        for i in indices:
            table = self.widget(i)
            if isinstance(table, QTableGroup):
                table = self.untileTable(i)
            tables.append(table)
        group = QTableGroup(tables, ori)

        for index in indices:
            self.replaceWidget(index, group.copy())

        self.setCurrentIndex(current_index)
        return None

    def tiledIndices(self, idx: int) -> list[int]:
        """Return indices of tables that are tiled with one at idx."""
        wdt = self.widget(idx)
        if isinstance(wdt, QTableGroup):
            wdt = cast(QTableGroup, wdt)
            return [self.tableIndex(table) for table in wdt.tables]
        else:
            return [idx]

    def replaceWidget(self, index: int, new: QtW.QWidget) -> None:
        """
        Replace table at index with new table.

        This function will NOT consider if the incoming and to-be-replaced tables
        are table group or not. This function is genuinely a method of QTabWidget.

        Parameters
        ----------
        index : int
            Index of the widget to be replaced.
        new : QWidget
            New widget to replace the old one.
        """
        text = self.tabText(index)
        self.removeTab(index)
        self.insertTab(index, new, text)
        return None

    def untileTable(self, index: int) -> QBaseTable:
        """Reset merge of the table group at index."""
        target_group = self.widget(index)
        if not isinstance(target_group, QTableGroup):
            return target_group

        current_index = self.currentIndex()
        target_group = cast(QTableGroup, target_group)
        n_merged = len(target_group.tables)
        appeared_idx: list[int] = []
        count = 0
        all_groups: list[QTableGroup] = []
        target_idx = -1
        for i in range(self.count()):
            wdt = self.widget(i)
            if target_group == wdt:
                all_groups.append(wdt)
                appeared_idx.append(i)
                if i == index:
                    target_idx = count
                count += 1

        unmerged = None
        for i, group in enumerate(all_groups):
            table = group.pop(target_idx)
            j = appeared_idx[i]
            if j == index:
                unmerged = table
                self.replaceWidget(j, table)
            elif n_merged == 2:
                self.replaceWidget(j, group.pop(0))

        if unmerged is None:
            raise RuntimeError("Unmerging could not be resolved.")
        self.setCurrentIndex(current_index)
        target_group.deleteLater()
        self.parentWidget().setCellFocus()
        return unmerged

    def copyData(self, index: int):
        """Copy all the data in the table at index to the clipboard."""
        table = self.tableAtIndex(index)
        h, w = table.dataShape()
        table.setSelections([(slice(0, h), slice(0, w))])
        table.copyToClipboard(headers=True)
        table.setSelections([])
        return None

    def openFinderDialog(self, index: int | None = None):
        """Open a dialog to find data in the table at index."""
        if index is not None:
            self.setCurrentIndex(index)
        ol = self._overlay
        ol.show()
        from ._finder import QFinderWidget

        _finder = ol.widget()
        if not isinstance(_finder, QFinderWidget):
            _finder = QFinderWidget(ol)
            _finder.searchBox().escClicked.connect(ol.hide)
            ol.addWidget(_finder)
            ol.setTitle("Find/Replace")
            _finder.searchBox().setFocus()
        return _finder

    def openFilterDialog(self, index: int | None = None):
        if index is not None:
            self.setCurrentIndex(index)
        ol = self._overlay
        ol.show()
        from ._eval import QLiteralFilterWidget

        _evaluator = QLiteralFilterWidget(ol)

        @_evaluator._line.escClicked.connect
        def _on_escape():
            ol.hide()
            self.parent().setCellFocus()

        ol.addWidget(_evaluator)
        ol.setTitle("Filter")
        _evaluator._line.setFocus()
        return _evaluator

    def openEvalDialog(self, index: int | None = None):
        if index is not None:
            self.setCurrentIndex(index)
        ol = self._overlay
        ol.show()
        from ._eval import QLiteralEvalWidget

        _evaluator = QLiteralEvalWidget(ol)

        @_evaluator._line.escClicked.connect
        def _on_escape():
            ol.hide()
            self.parent().setCellFocus()

        ol.addWidget(_evaluator)
        ol.setTitle("Evaluation")
        _evaluator._line.setFocus()
        return _evaluator

    def _group_index_to_tab_index(self, group: QTableGroup, index: int) -> int:
        # return the global in index of `index`-th table in `group`
        count = 0
        for i in range(self.count()):
            wdt = self.widget(i)
            if group == wdt:
                if count == index:
                    return i
                count += 1
        raise ValueError(f"{index} is not in {group}.")

    def _tab_index_to_group_index(self, index: int) -> int:
        wdt = cast(QTableGroup, self.widget(index))
        if isinstance(wdt, QTableGroup):
            count = 0
            for i in range(index):
                if wdt == self.widget(i):
                    count += 1
            return count
        else:
            raise ValueError(f"Widget at {index} is not a table group.")


class QEditabilityNotifier(QtW.QWidget):
    def __init__(self, table: QMutableTable, name: str) -> None:
        super().__init__()
        if len(name) > 20:
            name = name[:17] + "..."

        btn = QClickableLabel("Set editable (Ctrl+K, E)")
        btn.clicked.connect(self._on_button_clicked)
        self.setFont(QtGui.QFont("Arial"))

        _layout = QtW.QVBoxLayout()
        _layout.addWidget(_label(f"Table {name!r} is not editable."))
        _layout.addWidget(btn)

        self._table = weakref.ref(table)
        self.setLayout(_layout)

    def _on_button_clicked(self):
        table = self._table()
        if table is not None:
            table.setEditable(True)
            ol: QOverlayWidget = self.parentWidget()
            ol.setVisible(False)
        return None


class QLatestVersionNotifier(QtW.QWidget):
    def __init__(self, version: str):
        super().__init__()

        btn = QClickableLabel("See release note")
        btn.clicked.connect(self.open_release_note)
        self.setFont(QtGui.QFont("Arial"))

        _layout = QtW.QVBoxLayout()
        _layout.addWidget(QtW.QLabel(f"Latest version {version} is available."))
        _layout.addWidget(btn)

        self.setLayout(_layout)

    def open_release_note(self):
        """Open the release note page in the default browser."""
        from qtpy.QtGui import QDesktopServices
        from qtpy.QtCore import QUrl

        url = "https://github.com/hanjinliu/tabulous/releases"
        QDesktopServices.openUrl(QUrl(url))
        return None


class QNotSavedNotifier(QtW.QWidget):
    def __init__(self, name: str):
        super().__init__()
        btn_save = QClickableLabel("Save to the source file")
        btn_save_new = QClickableLabel("Save to a new file")
        btn_delete = QClickableLabel("Discard changes")
        btn_cancel = QClickableLabel("Cancel")

        _layout = QtW.QVBoxLayout()
        _layout.addWidget(_label(f"You may have unsaved changes in table {name!r}."))
        _layout.addWidget(btn_save)
        _layout.addWidget(btn_save_new)
        _layout.addWidget(btn_delete)
        _layout.addWidget(btn_cancel)

        btn_save.clicked.connect(self._on_save_clicked)
        btn_save_new.clicked.connect(self._on_save_new_clicked)
        btn_delete.clicked.connect(self._on_delete_clicked)
        btn_cancel.clicked.connect(self._on_cancel_clicked)

        self._name = name
        self.setLayout(_layout)

    def parentViewer(self) -> _QtMainWidgetBase:
        from tabulous._qt._mainwindow import _QtMainWidgetBase

        parent = self.parentWidget()
        while not isinstance(parent, _QtMainWidgetBase):
            parent = parent.parentWidget()
            if parent is None:
                raise RuntimeError("Cannot find the viewer.")
        return parent

    def _on_save_clicked(self):
        qviewer = self.parentViewer()
        table = qviewer._table_viewer.tables[self._name]
        table.save(table.source.path)
        self._on_delete_clicked()

    def _on_save_new_clicked(self):
        qviewer = self.parentViewer()
        table = qviewer._table_viewer.tables[self._name]
        self._on_cancel_clicked()
        path = qviewer._table_viewer.history_manager.openFileDialog(
            mode="w", caption="Save table"
        )
        if path:
            table.save(path)
            del qviewer._table_viewer.tables[self._name]

    def _on_delete_clicked(self):
        qviewer = self.parentViewer()
        del qviewer._table_viewer.tables[self._name]
        self._on_cancel_clicked()

    def _on_cancel_clicked(self):
        ol: QOverlayWidget = self.parentWidget()
        ol.setVisible(False)


def _label(text: str) -> QtW.QLabel:
    w = QtW.QLabel(text)
    w.setFont(QtGui.QFont("Arial", 11))
    w.setContentsMargins(0, 0, 0, 8)
    return w



from __future__ import annotations
from typing import TYPE_CHECKING, cast
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt

from ._table_base import QBaseTable
from ._line_edit import QCellLiteralEdit

from tabulous._magicgui._timedelta import QTimeDeltaEdit

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
    from pandas.core.dtypes.dtypes import CategoricalDtype
    from ._enhanced_table import _QTableViewEnhanced
    from ._item_model import AbstractDataFrameModel

HOV_COLOR = QtGui.QColor(75, 75, 242, 80)


class TableItemDelegate(QtW.QStyledItemDelegate):
    """Displays table widget items with properly formatted numbers."""

    def createEditor(
        self, parent: QtW.QWidget, option, index: QtCore.QModelIndex
    ) -> QtW.QWidget:
        """Create different type of editors for different dtypes."""
        qtable_view: _QTableViewEnhanced = parent.parent()
        if qtable_view.model()._editable:
            model = qtable_view.model()
            df = model.df
            row = index.row()
            col = index.column()
            font = QtGui.QFont(
                qtable_view._font, int(qtable_view._font_size * qtable_view.zoom())
            )
            if row >= df.shape[0] or col >= df.shape[1]:
                # out-of-bounds
                line = qtable_view._create_eval_editor(moveto=(row, col))
                line.setFont(font)
                # NOTE: Don't return the line edit, otherwise it closes when outside
                # the cell is clicked!
                return None

            dtype: np.dtype = df.dtypes.values[col]
            value = df.iat[row, col]
            if dtype == "category":
                # use combobox for categorical data
                dtype: CategoricalDtype
                cbox = QtW.QComboBox(parent)
                cbox.setFont(font)
                choices = list(map(str, dtype.categories))
                cbox.addItems(choices)
                cbox.setCurrentIndex(choices.index(str(value)))
                cbox.currentIndexChanged.connect(qtable_view.setFocus)
                return cbox
            elif dtype == "bool":
                # use checkbox for boolean data
                cbox = QtW.QComboBox(parent)
                cbox.setFont(font)
                choices = ["True", "False"]
                cbox.addItems(choices)
                cbox.setCurrentIndex(0 if value else 1)
                cbox.currentIndexChanged.connect(qtable_view.setFocus)
                return cbox
            elif dtype.kind == "M":
                dt = QtW.QDateTimeEdit(parent)
                dt.setFont(font)
                dt.setButtonSymbols(QtW.QAbstractSpinBox.ButtonSymbols.NoButtons)
                value: pd.Timestamp
                dt.setDateTime(value.to_pydatetime())
                return dt
            elif dtype.kind == "m":
                td = QTimeDeltaEdit(parent)
                td.setFont(font)
                td.setButtonSymbols(QtW.QAbstractSpinBox.ButtonSymbols.NoButtons)
                td.setValue(value)
                return td
            else:
                line = qtable_view._create_eval_editor(moveto=(row, col))
                line.setFont(font)
                return None

    def setEditorData(self, editor: QtW.QWidget, index: QtCore.QModelIndex) -> None:
        super().setEditorData(editor, index)
        if isinstance(editor, QtW.QComboBox):
            editor = cast(QtW.QComboBox, editor)
            editor.showPopup()
        elif isinstance(editor, QCellLiteralEdit):
            editor = cast(QCellLiteralEdit, editor)
            editor._on_text_changed(editor.text())
        return None

    def setModelData(
        self,
        editor: QtW.QWidget,
        model: AbstractDataFrameModel,
        index: QtCore.QModelIndex,
    ) -> None:
        if isinstance(editor, QtW.QDateTimeEdit):
            editor = cast(QtW.QDateTimeEdit, editor)
            dt = editor.dateTime().toPyDateTime()
            model.setData(index, dt, Qt.ItemDataRole.EditRole)
            return None
        elif isinstance(editor, QCellLiteralEdit):
            editor = cast(QCellLiteralEdit, editor)
            if editor.mode is QCellLiteralEdit.Mode.EVAL:
                # set the evaluated data
                return None
        elif isinstance(editor, QTimeDeltaEdit):
            editor = cast(QTimeDeltaEdit, editor)
            model.setData(index, editor.value(), Qt.ItemDataRole.EditRole)
            return None

        return super().setModelData(editor, model, index)

    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtW.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ):
        option.textElideMode = Qt.TextElideMode.ElideNone
        super().paint(painter, option, index)
        if option.state & QtW.QStyle.StateFlag.State_MouseOver:
            painter.setPen(QtGui.QPen(HOV_COLOR, 2))
            option.rect.adjust(1, 1, -1, -1)
            painter.drawRect(option.rect)

    def initStyleOption(
        self, option: QtW.QStyleOptionViewItem, index: QtCore.QModelIndex
    ):
        super().initStyleOption(option, index)
        if option.state & QtW.QStyle.StateFlag.State_HasFocus:
            option.state = option.state & ~QtW.QStyle.StateFlag.State_HasFocus

    def parentTable(self) -> QBaseTable:
        return self.parent().parent()



from __future__ import annotations
import sys
from typing import Any, Callable
from types import TracebackType

__all__ = ["SelectionRangeError", "ExceptionHandler"]


class SelectionRangeError(ValueError):
    """Raised when the exception is caused by wrong selection range(s)."""


class TableImmutableError(ValueError):
    """Raised when immutable table was being tried to edit."""


class TriggerParent(RuntimeError):
    """Should try to trigger parent's keybinding instead of the current one"""


class TableNotOrderedError(ValueError):
    """Raised when an operation is undefined if table is not ordered."""


class UnreachableError(RuntimeError):
    """Raised when an unreachable code is reached."""


class ExceptionHandler:
    """Handle exceptions in the GUI thread."""

    def __init__(
        self, hook: Callable[[type[Exception], Exception, TracebackType], Any]
    ):
        self._excepthook = hook

    def __enter__(self):
        self._original_excepthook = sys.excepthook
        sys.excepthook = self._excepthook
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        sys.excepthook = self._original_excepthook
        return None



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Union
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Signal

from magicgui import __version__ as MAGICGUI_VERSION, register_type

if int(MAGICGUI_VERSION.split(".")[1]) >= 7:
    from magicgui.widgets.bases import ValueWidget
    from magicgui.types import Undefined
else:
    from magicgui.widgets._bases import ValueWidget

    Undefined = None

from magicgui.backends._qtpy.widgets import QBaseValueWidget

_TimeDeltaLike = Union[datetime.timedelta, str, int, float]


class Section:
    NONE = -1
    DAY = 0
    HOUR = 1
    MINUTE = 2
    SECOND = 3


class QTimeDeltaEdit(QtW.QAbstractSpinBox):
    valueChanged = Signal(datetime.timedelta)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._value = datetime.timedelta(0)
        self._min = datetime.timedelta(-999999999)
        self._max = datetime.timedelta(999999999)
        self._step = datetime.timedelta(seconds=1)
        self.setLineEdit(_QTimeDeltaLineEdit(self._value, self))
        self.lineEdit().timedeltaChanged.connect(self.setValue)
        self.lineEdit().sectionChanged.connect(self._on_section_changed)
        self.setMinimumSize(140, 20)

    if TYPE_CHECKING:

        def lineEdit(self) -> _QTimeDeltaLineEdit:
            ...

    def value(self) -> datetime.timedelta:
        return self._value

    def setValue(self, value: _TimeDeltaLike) -> None:
        self._value = to_timedelta(value)
        if self._min <= self._value <= self._max:
            self._update_text(self._value)
            self.update()
            self.valueChanged.emit(value)
            return None
        raise ValueError(f"{value} is out of bound.")

    def singleStep(self) -> datetime.timedelta:
        return self._step

    def setSingleStep(self, step: _TimeDeltaLike):
        self._step = to_timedelta(step)

    def stepBy(self, steps: int) -> None:
        """Increment/decrement the time delta by the step value."""
        line = self.lineEdit()
        start = line.selectionStart()
        length = line.selectionLength()
        self.setValue(self.value() + steps * self.singleStep())
        line.setSelection(max(start, 0), length)

    def maximum(self) -> datetime.timedelta:
        """The maximum time delta value."""
        return self._max

    def setMaximum(self, max: _TimeDeltaLike) -> None:
        """Set the maximum time delta value."""
        self._max = to_timedelta(max)

    def minimum(self) -> datetime.timedelta:
        """The minimum time delta value."""
        return self._min

    def setMinimum(self, min: _TimeDeltaLike) -> None:
        """Set the minimum time delta value."""
        self._min = to_timedelta(min)

    def setRange(self, min: _TimeDeltaLike, max: _TimeDeltaLike):
        """Set the minimum and the maximum time delta value."""
        self.setMinimum(min)
        self.setMaximum(max)

    def stepEnabled(self):
        flags = QtW.QAbstractSpinBox.StepEnabledFlag.StepNone
        if self.isReadOnly():
            return flags
        if self._value < self._max:
            flags |= QtW.QAbstractSpinBox.StepEnabledFlag.StepUpEnabled
        if self._value > self._min:
            flags |= QtW.QAbstractSpinBox.StepEnabledFlag.StepDownEnabled
        return flags

    def stepType(self):
        return QtW.QAbstractSpinBox.StepType.DefaultStepType

    def sizeHint(self):
        # copied from superqt.QLargeIntSpinBox
        self.ensurePolished()
        fm = QtGui.QFontMetrics(self.font())
        h = self.lineEdit().sizeHint().height()
        if hasattr(fm, "horizontalAdvance"):
            # Qt >= 5.11
            w = fm.horizontalAdvance(str(self._value)) + 3
        else:
            w = fm.width(str(self._value)) + 3
        w = max(36, w)
        opt = QtW.QStyleOptionSpinBox()
        self.initStyleOption(opt)
        hint = QtCore.QSize(w, h)
        return self.style().sizeFromContents(
            QtW.QStyle.ContentsType.CT_SpinBox, opt, hint, self
        )

    def _update_text(self, value: datetime.timedelta) -> None:
        self.lineEdit().setTimedelta(value, emit=False)

    def _on_section_changed(self, section: int) -> None:
        if section == Section.SECOND:
            self.setSingleStep(datetime.timedelta(seconds=1))
        elif section == Section.MINUTE:
            self.setSingleStep(datetime.timedelta(minutes=1))
        elif section == Section.HOUR:
            self.setSingleStep(datetime.timedelta(hours=1))
        elif section == Section.DAY:
            self.setSingleStep(datetime.timedelta(days=1))
        self.lineEdit()._select_section(section)


class _QTimeDeltaLineEdit(QtW.QLineEdit):
    sectionChanged = Signal(int)  # position
    timedeltaChanged = Signal(datetime.timedelta)

    def __init__(self, value: datetime.timedelta, parent=None):
        super().__init__(parent)
        self.setTimedelta(value)
        # "000 days 00:00:00" format
        self.setInputMask(r"###0\ \d\ays\ 00\:00\:00")
        self.cursorPositionChanged.connect(self._on_cursor_position_changed)

    def timedelta(self) -> datetime.timedelta:
        days, rest = self.text().split("days")
        hours, mins, secs = rest.split(":")
        days, hours, mins, secs = map(int, (days, hours, mins, secs))
        return datetime.timedelta(days, hours * 3600 + mins * 60 + secs)

    def setTimedelta(self, value: datetime.timedelta, emit=True) -> None:
        days = value.days
        sec_tot = value.seconds
        hours, min_rest = divmod(sec_tot, 3600)
        mins, secs = divmod(min_rest, 60)
        text = f"{days: 3} days {hours:02}:{mins:02}:{secs:02}"
        self.setText(text)
        if emit:
            self.timedeltaChanged.emit(value)

    def sectionUnderCursor(self, pos: int) -> int:
        # nchar = 18
        if 16 <= pos:
            return Section.SECOND
        elif 13 <= pos:
            return Section.MINUTE
        elif 10 <= pos:
            return Section.HOUR
        else:
            return Section.DAY

    def _select_section(self, section: int) -> None:
        if section == Section.SECOND:
            self.setSelection(16, 2)
        elif section == Section.MINUTE:
            self.setSelection(13, 2)
        elif section == Section.HOUR:
            self.setSelection(10, 2)
        else:
            self.setSelection(1, 3)

    def _on_cursor_position_changed(self, old: int, new: int) -> None:
        section_old = self.sectionUnderCursor(old)
        section_new = self.sectionUnderCursor(new)
        if section_old != section_new:
            self.sectionChanged.emit(section_new)


def to_timedelta(value: _TimeDeltaLike) -> datetime.timedelta:
    if not isinstance(value, datetime.timedelta):
        import pandas as pd

        pd_td = pd.to_timedelta(value)
        if isinstance(pd_td, pd.Timedelta):
            value = pd_td.to_pytimedelta()
        else:
            raise TypeError(f"Cannot convert {value} to timedelta")
    return value


class _TimeDeltaEdit(QBaseValueWidget):
    def __init__(self, **kwargs):
        super().__init__(QTimeDeltaEdit, "value", "setValue", "valueChanged", **kwargs)


class TimeDeltaEdit(ValueWidget):
    def __init__(self, value=Undefined, **kwargs) -> None:
        if value is Undefined:
            value = datetime.timedelta(seconds=0)
        super().__init__(
            widget_type=_TimeDeltaEdit,
            value=value,
            **kwargs,
        )


register_type(datetime.timedelta, widget_type=TimeDeltaEdit)



from __future__ import annotations
from typing import Callable, TYPE_CHECKING

if TYPE_CHECKING:
    import pandas as pd

_SAMPLE_PROVIDERS: dict[str, Callable[[str], pd.DataFrame]] = {}


def register_sample_plugin(
    provider: Callable[[str], pd.DataFrame] | None = None,
    plugin_name: str | None = None,
):
    def _register(func):
        nonlocal plugin_name
        if plugin_name is None:
            plugin_name = func.__name__
        _SAMPLE_PROVIDERS[plugin_name] = func
        return func

    if provider is None:
        return _register
    else:
        return _register(provider)


@register_sample_plugin
def seaborn(name: str):
    import seaborn as sns

    return sns.load_dataset(name)


def open_sample(name: str, plugin_name: str) -> pd.DataFrame:
    provider = _SAMPLE_PROVIDERS.get(plugin_name)
    if provider is None:
        raise ValueError(f"No plugin named {plugin_name}")
    return provider(name)



from __future__ import annotations

from qtpy import QtWidgets as QtW

from tabulous._qt._mainwindow import QMainWindow
from tabulous._utils import get_config, update_config
from tabulous._magicgui._toggle_switch import QToggleSwitch


class QGeneralPanel(QtW.QWidget):
    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        self._setup_ui()

    def get_updated_config(self):
        cfg = get_config()
        cfg.window.ask_on_close = self._ask_on_close.isChecked()
        cfg.window.notify_latest = self._notify_latest.isChecked()
        cfg.window.selection_editor = self._selection_editor.isChecked()
        cfg.window.show_console = self._show_console.isChecked()
        cfg.console_namespace.load_startup_file = self._load_startup.isChecked()
        return cfg

    def _update_config(self):
        cfg = self.get_updated_config()
        update_config(cfg)
        QMainWindow.reload_config()

    def _setup_ui(self):
        from tabulous._utils import get_config

        cfg = get_config()

        self._ask_on_close = QToggleSwitch()
        self._ask_on_close.setChecked(cfg.window.ask_on_close)

        self._notify_latest = QToggleSwitch()
        self._notify_latest.setChecked(cfg.window.notify_latest)

        self._selection_editor = QToggleSwitch()
        self._selection_editor.setChecked(cfg.window.selection_editor)

        self._show_console = QToggleSwitch()
        self._show_console.setChecked(cfg.window.show_console)

        self._load_startup = QToggleSwitch()
        self._load_startup.setChecked(cfg.console_namespace.load_startup_file)

        _layout = QtW.QFormLayout()

        _layout.addRow("Ask before closing window", self._ask_on_close)
        _layout.addRow("Notify latest version", self._notify_latest)
        _layout.addRow("Show selection editor", self._selection_editor)
        _layout.addRow("Show Qt console", self._show_console)
        _layout.addRow("Load IPython startup files", self._load_startup)

        self._ask_on_close.toggled.connect(self._update_config)
        self._notify_latest.toggled.connect(self._update_config)
        self._selection_editor.toggled.connect(self._update_config)
        self._show_console.toggled.connect(self._update_config)
        self._load_startup.toggled.connect(self._update_config)

        self.setLayout(_layout)



from __future__ import annotations
from typing import Any, NamedTuple
from functools import lru_cache
import colorsys
from tabulous.types import ColorType, ColorMapping

import numpy as np


class ColorTuple(NamedTuple):
    """8-bit color tuple."""

    r: int
    g: int
    b: int
    a: int = 255

    @property
    def opacity(self) -> float:
        """Return the opacity as a float between 0 and 1."""
        return self.a / 255.0

    @property
    def html(self) -> str:
        """Return a HTML color string."""
        if self.a == 255:
            return f"#{self.r:02X}{self.g:02X}{self.b:02X}"
        return f"#{self.r:02X}{self.g:02X}{self.b:02X}{self.a:02X}"

    @property
    def hlsa(self) -> tuple[float, float, float, float]:
        """Return the color as HSLA."""
        hlsa_float = colorsys.rgb_to_hls(
            self.r / 255.0, self.g / 255.0, self.b / 255.0
        ) + (self.opacity,)
        return tuple(int(round(c * 255)) for c in hlsa_float)

    @property
    def hsva(self) -> tuple[float, float, float, float]:
        """Return the color as HSVA."""
        hsva_float = colorsys.rgb_to_hsv(
            self.r / 255.0, self.g / 255.0, self.b / 255.0
        ) + (self.opacity,)
        return tuple(int(round(c * 255)) for c in hsva_float)

    @classmethod
    def from_html(cls, html: str) -> ColorTuple:
        """Create a ColorTuple from a HTML color string."""
        if html.startswith("#"):
            html = html[1:]
        if len(html) == 6:
            html += "FF"
        return cls(*[int(html[i : i + 2], 16) for i in range(0, 8, 2)])

    @classmethod
    def from_hlsa(cls, *hlsa) -> ColorTuple:
        """Create a ColorTuple from HSLA."""
        if len(hlsa) == 1:
            hlsa = hlsa[0]
        if len(hlsa) == 3:
            hls = hlsa
            alpha = 255
        hls = tuple(c / 255.0 for c in hls)
        return cls(*[int(round(c * 255)) for c in colorsys.hls_to_rgb(*hls)], alpha)

    @classmethod
    def from_hsva(cls, *hsva) -> ColorTuple:
        """Create a ColorTuple from HSVA."""
        if len(hsva) == 1:
            hsva = hsva[0]
        if len(hsva) == 3:
            hsv = hsva
            alpha = 255
        hsv_float = tuple(c / 255.0 for c in hsv)
        return cls(
            *[int(round(c * 255)) for c in colorsys.hsv_to_rgb(*hsv_float)], alpha
        )

    def equals(self, other):
        other = normalize_color(other)
        return self == other

    def brighten(self, ratio: float) -> ColorTuple:
        """Set the saturation of the color."""
        hsv = self.hsva[:3]
        val = round(hsv[2] * (1 + ratio))
        val = min(255, max(0, val))
        hsv = (hsv[0], hsv[1], val)
        return ColorTuple.from_hsva(hsv)

    def mix(self, other, ratio: float = 0.5, alpha: bool = False) -> ColorTuple:
        """Mix the color with another color."""
        other = normalize_color(other)
        if alpha:
            _alpha = _8bit(self.a * (1 - ratio) + other.a * ratio)
        else:
            _alpha = self.a
        return ColorTuple(
            _8bit(self.r * (1 - ratio) + other.r * ratio),
            _8bit(self.g * (1 - ratio) + other.g * ratio),
            _8bit(self.b * (1 - ratio) + other.b * ratio),
            _alpha,
        )


def _8bit(x: float) -> int:
    return max(min(int(round(x)), 255), 0)


def normalize_color(color: ColorType) -> ColorTuple:
    """Normalize a color-like object to a ColorTuple."""
    if isinstance(color, str):
        return ColorTuple(*_str_color_to_tuple(color))
    if hasattr(color, "__iter__"):
        out = [int(c) for c in color]
        if len(out) == 3:
            out += [255]
        elif len(out) == 4:
            pass
        else:
            raise ValueError(f"Invalid color: {color!r}")
        return ColorTuple(*out)
    raise ValueError(f"Invalid color: {color!r}")


def rgba_to_str(rgba: tuple[int, int, int, int]) -> str:
    color_name = COLORS_BY_VALUE.get(rgba, None)
    if color_name is None:
        code = "#" + "".join(hex(c)[2:].upper().zfill(2) for c in rgba)
        if code.endswith("FF"):
            code = code[:-2]
        return code
    return color_name


class ConvertedColormap:
    def __init__(self, func: ColorMapping):
        self.func = func
        self.__name__ = f"{type(self).__name__}<{func.__name__}>"
        self.__annotations__ = func.__annotations__

    def __repr__(self):
        return f"{type(self).__name__}<{self.func!r}>"


class InvertedColormap(ConvertedColormap):
    @classmethod
    def from_colormap(cls, cmap: ColorMapping) -> ColorMapping:
        """Convert a colormap into return an inverted one."""
        if isinstance(cmap, cls):
            return cmap.func
        return cls(cmap)

    def __call__(self, x: Any) -> ColorType:
        color = self.func(x)
        if color is None:
            return color
        color = np.array(normalize_color(color), dtype=np.uint8)
        color[:3] = 255 - color[:3]
        return color


class OpacityColormap(ConvertedColormap):
    def __init__(self, func: ColorMapping, opacity: float):
        super().__init__(func)
        if opacity < 0 or 1 < opacity:
            raise ValueError(f"Opacity must be between 0 and 1, got {opacity}")
        self._alpha = int(opacity * 255)

    @classmethod
    def from_colormap(cls, cmap: ColorMapping, opacity: float) -> ColorMapping:
        """Convert a colormap into an new one with given alpha channel."""
        if isinstance(cmap, cls):
            return cls(cmap.func, opacity)
        return cls(cmap, opacity)

    def __call__(self, x: Any) -> ColorType:
        color = self.func(x)
        if color is None:
            return color
        color = np.array(normalize_color(color), dtype=np.uint8)
        color[3] = self._alpha
        return color


class BrightenedColormap(ConvertedColormap):
    def __init__(self, func: ColorMapping, factor: float):
        super().__init__(func)
        if factor < -1:
            raise ValueError(f"Brightening factor fell below -1.0: {factor}")
        if 1 < factor:
            raise ValueError(f"Brightening factor exceeded 1.0: {factor}")
        self._factor = factor

    @classmethod
    def from_colormap(cls, cmap: ColorMapping, factor: float) -> ColorMapping:
        """Convert a colormap into an new one with given brightening factor."""
        if isinstance(cmap, cls):
            return cls(cmap.func, cmap._factor + factor)
        return cls(cmap, factor)

    def __call__(self, x: Any) -> ColorType:
        color = self.func(x)
        if color is None:
            return color
        color = np.array(normalize_color(color), dtype=np.float64)
        factor = self._factor
        if factor > 0:
            extreme = np.array([255, 255, 255, 255], dtype=np.float64)
        else:
            extreme = np.array([0, 0, 0, 255], dtype=np.float64)
        color = color * (1 - factor) + extreme * factor
        return np.round(color).astype(np.uint8)


@lru_cache(maxsize=64)
def _str_color_to_tuple(color: str) -> tuple[int, int, int, int]:
    out = COLORS_BY_NAME.get(color, None)
    if out is not None:
        return out
    if color.startswith("#"):
        color = color[1:]
    if len(color) == 6:
        return int(color[:2], 16), int(color[2:4], 16), int(color[4:6], 16), 255
    elif len(color) == 8:
        return (
            int(color[:2], 16),
            int(color[2:4], 16),
            int(color[4:6], 16),
            int(color[6:8], 16),
        )
    raise ValueError(f"Invalid color: {color!r}")


# This dict is modified from pydantic (MIT licence)
# See https://github.com/samuelcolvin/pydantic
COLORS_BY_NAME = {
    "aliceblue": (240, 248, 255, 255),
    "antiquewhite": (250, 235, 215, 255),
    "aqua": (0, 255, 255, 255),
    "aquamarine": (127, 255, 212, 255),
    "azure": (240, 255, 255, 255),
    "beige": (245, 245, 220, 255),
    "bisque": (255, 228, 196, 255),
    "black": (0, 0, 0, 255),
    "blanchedalmond": (255, 235, 205, 255),
    "blue": (0, 0, 255, 255),
    "blueviolet": (138, 43, 226, 255),
    "brown": (165, 42, 42, 255),
    "burlywood": (222, 184, 135, 255),
    "cadetblue": (95, 158, 160, 255),
    "chartreuse": (127, 255, 0, 255),
    "chocolate": (210, 105, 30, 255),
    "coral": (255, 127, 80, 255),
    "cornflowerblue": (100, 149, 237, 255),
    "cornsilk": (255, 248, 220, 255),
    "crimson": (220, 20, 60, 255),
    "cyan": (0, 255, 255, 255),
    "darkblue": (0, 0, 139, 255),
    "darkcyan": (0, 139, 139, 255),
    "darkgoldenrod": (184, 134, 11, 255),
    "darkgray": (169, 169, 169, 255),
    "darkgreen": (0, 100, 0, 255),
    "darkgrey": (169, 169, 169, 255),
    "darkkhaki": (189, 183, 107, 255),
    "darkmagenta": (139, 0, 139, 255),
    "darkolivegreen": (85, 107, 47, 255),
    "darkorange": (255, 140, 0, 255),
    "darkorchid": (153, 50, 204, 255),
    "darkred": (139, 0, 0, 255),
    "darksalmon": (233, 150, 122, 255),
    "darkseagreen": (143, 188, 143, 255),
    "darkslateblue": (72, 61, 139, 255),
    "darkslategray": (47, 79, 79, 255),
    "darkslategrey": (47, 79, 79, 255),
    "darkturquoise": (0, 206, 209, 255),
    "darkviolet": (148, 0, 211, 255),
    "deeppink": (255, 20, 147, 255),
    "deepskyblue": (0, 191, 255, 255),
    "dimgray": (105, 105, 105, 255),
    "dimgrey": (105, 105, 105, 255),
    "dodgerblue": (30, 144, 255, 255),
    "firebrick": (178, 34, 34, 255),
    "floralwhite": (255, 250, 240, 255),
    "forestgreen": (34, 139, 34, 255),
    "fuchsia": (255, 0, 255, 255),
    "gainsboro": (220, 220, 220, 255),
    "ghostwhite": (248, 248, 255, 255),
    "gold": (255, 215, 0, 255),
    "goldenrod": (218, 165, 32, 255),
    "gray": (128, 128, 128, 255),
    "green": (0, 128, 0, 255),
    "greenyellow": (173, 255, 47, 255),
    "grey": (128, 128, 128, 255),
    "honeydew": (240, 255, 240, 255),
    "hotpink": (255, 105, 180, 255),
    "indianred": (205, 92, 92, 255),
    "indigo": (75, 0, 130, 255),
    "ivory": (255, 255, 240, 255),
    "khaki": (240, 230, 140, 255),
    "lavender": (230, 230, 250, 255),
    "lavenderblush": (255, 240, 245, 255),
    "lawngreen": (124, 252, 0, 255),
    "lemonchiffon": (255, 250, 205, 255),
    "lightblue": (173, 216, 230, 255),
    "lightcoral": (240, 128, 128, 255),
    "lightcyan": (224, 255, 255, 255),
    "lightgoldenrodyellow": (250, 250, 210, 255),
    "lightgray": (211, 211, 211, 255),
    "lightgreen": (144, 238, 144, 255),
    "lightgrey": (211, 211, 211, 255),
    "lightpink": (255, 182, 193, 255),
    "lightsalmon": (255, 160, 122, 255),
    "lightseagreen": (32, 178, 170, 255),
    "lightskyblue": (135, 206, 250, 255),
    "lightslategray": (119, 136, 153, 255),
    "lightslategrey": (119, 136, 153, 255),
    "lightsteelblue": (176, 196, 222, 255),
    "lightyellow": (255, 255, 224, 255),
    "lime": (0, 255, 0, 255),
    "limegreen": (50, 205, 50, 255),
    "linen": (250, 240, 230, 255),
    "magenta": (255, 0, 255, 255),
    "maroon": (128, 0, 0, 255),
    "mediumaquamarine": (102, 205, 170, 255),
    "mediumblue": (0, 0, 205, 255),
    "mediumorchid": (186, 85, 211, 255),
    "mediumpurple": (147, 112, 219, 255),
    "mediumseagreen": (60, 179, 113, 255),
    "mediumslateblue": (123, 104, 238, 255),
    "mediumspringgreen": (0, 250, 154, 255),
    "mediumturquoise": (72, 209, 204, 255),
    "mediumvioletred": (199, 21, 133, 255),
    "midnightblue": (25, 25, 112, 255),
    "mintcream": (245, 255, 250, 255),
    "mistyrose": (255, 228, 225, 255),
    "moccasin": (255, 228, 181, 255),
    "navajowhite": (255, 222, 173, 255),
    "navy": (0, 0, 128, 255),
    "oldlace": (253, 245, 230, 255),
    "olive": (128, 128, 0, 255),
    "olivedrab": (107, 142, 35, 255),
    "orange": (255, 165, 0, 255),
    "orangered": (255, 69, 0, 255),
    "orchid": (218, 112, 214, 255),
    "palegoldenrod": (238, 232, 170, 255),
    "palegreen": (152, 251, 152, 255),
    "paleturquoise": (175, 238, 238, 255),
    "palevioletred": (219, 112, 147, 255),
    "papayawhip": (255, 239, 213, 255),
    "peachpuff": (255, 218, 185, 255),
    "peru": (205, 133, 63, 255),
    "pink": (255, 192, 203, 255),
    "plum": (221, 160, 221, 255),
    "powderblue": (176, 224, 230, 255),
    "purple": (128, 0, 128, 255),
    "red": (255, 0, 0, 255),
    "rosybrown": (188, 143, 143, 255),
    "royalblue": (65, 105, 225, 255),
    "saddlebrown": (139, 69, 19, 255),
    "salmon": (250, 128, 114, 255),
    "sandybrown": (244, 164, 96, 255),
    "seagreen": (46, 139, 87, 255),
    "seashell": (255, 245, 238, 255),
    "sienna": (160, 82, 45, 255),
    "silver": (192, 192, 192, 255),
    "skyblue": (135, 206, 235, 255),
    "slateblue": (106, 90, 205, 255),
    "slategray": (112, 128, 144, 255),
    "slategrey": (112, 128, 144, 255),
    "snow": (255, 250, 250, 255),
    "springgreen": (0, 255, 127, 255),
    "steelblue": (70, 130, 180, 255),
    "tan": (210, 180, 140, 255),
    "teal": (0, 128, 128, 255),
    "thistle": (216, 191, 216, 255),
    "tomato": (255, 99, 71, 255),
    "turquoise": (64, 224, 208, 255),
    "violet": (238, 130, 238, 255),
    "wheat": (245, 222, 179, 255),
    "white": (255, 255, 255, 255),
    "whitesmoke": (245, 245, 245, 255),
    "yellow": (255, 255, 0, 255),
    "yellowgreen": (154, 205, 50, 255),
}

COLORS_BY_VALUE = {v: k for k, v in COLORS_BY_NAME.items()}



from __future__ import annotations

from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt

if TYPE_CHECKING:
    from ..._table_stack import QTabbedTableStack
    from .._base import QBaseTable


class QOverlayFrame(QtW.QDialog):
    """The overlay frame widget used to display the overlaid widget in tables."""

    def __init__(
        self,
        content: QtW.QWidget,
        viewport: QtW.QWidget,
        grip: bool = True,
        drag: bool = True,
    ):
        super().__init__(viewport, Qt.WindowType.SubWindow)

        self.setLayout(QtW.QVBoxLayout())

        self._label_widget = QtW.QLabel()
        self._label_widget.setContentsMargins(5, 2, 5, 2)
        self._drag_start = None

        content.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )

        if grip:
            self._add_header()

        self.layout().addWidget(content)
        self.layout().setSpacing(0)
        self.setContentsMargins(0, 0, 0, 0)

        if grip:
            self.layout().setContentsMargins(0, 0, 0, 0)
            self._add_footer()
        else:
            self.layout().setContentsMargins(2, 2, 2, 2)

        self.setDraggable(drag)

    def label(self) -> str:
        """The label of the overlay widget."""
        return self._label_widget.text()

    def setLabel(self, label: str) -> None:
        """Set the label of the overlay widget in the bottom."""
        return self._label_widget.setText(label)

    def draggable(self) -> bool:
        """Return if the overlay widget is draggable."""
        return self._draggable

    def setDraggable(self, draggable: bool) -> None:
        """Set if the overlay widget is draggable."""
        self._draggable = draggable
        if draggable:
            effect = QtW.QGraphicsDropShadowEffect(
                self, blurRadius=5, xOffset=2, yOffset=3
            )
            effect.setColor(QtGui.QColor(128, 128, 128, 72))
            self.setGraphicsEffect(effect)
        else:
            self.setGraphicsEffect(None)
        return None

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        """Change the z-order of the overlay widget."""
        if self._draggable:
            self._drag_start = event.pos()
        self.raise_()
        self.setFocus()
        return None

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        self._drag_start = None
        return None

    def keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if a0.key() == Qt.Key.Key_Escape:
            # by default, this will close the overlay
            a0.ignore()
            return None
        return super().keyPressEvent(a0)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        if self._drag_start is not None:
            self.move(self.mapToParent(event.pos() - self._drag_start))
        return None

    def parentTable(self) -> QBaseTable:
        return self.parent().parent().parent()

    def tableStack(self) -> QTabbedTableStack:
        """The parent table stack."""
        return self.parentTable().tableStack()

    def _add_header(self):
        self.layout().addWidget(
            QtW.QSizeGrip(self),
            False,
            Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop,
        )

    def _add_footer(self):
        _footer = QtW.QWidget()
        _footer.setLayout(QtW.QHBoxLayout())
        _footer.layout().addWidget(
            self._label_widget, False, Qt.AlignmentFlag.AlignLeft
        )
        _footer.layout().addWidget(
            QtW.QSizeGrip(self),
            False,
            Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignBottom,
        )
        _footer.setContentsMargins(0, 0, 0, 0)
        _footer.layout().setContentsMargins(0, 0, 0, 0)
        _footer.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Minimum
        )
        self.layout().addWidget(_footer)



from __future__ import annotations
from typing import Any
import pandas as pd
from collections_undo import arguments

from ._base import QMutableSimpleTable, DataFrameModel
from tabulous._dtype import get_converter


class QTableLayer(QMutableSimpleTable):
    def getDataFrame(self) -> pd.DataFrame:
        return self._data_raw

    @QMutableSimpleTable._mgr.interface
    def setDataFrame(self, data: pd.DataFrame) -> None:
        self._data_raw = data
        self.model().df = data
        self.setProxy(None)
        self._qtable_view.viewport().update()
        return

    @setDataFrame.server
    def setDataFrame(self, data):
        try:
            return arguments(self.getDataFrame())
        except Exception:
            return None

    @setDataFrame.set_formatter
    def _setDataFrame_fmt(self, data: pd.DataFrame):
        return f"set new data of shape {data.shape}"

    __delete = object()

    @QMutableSimpleTable._mgr.interface
    def assignColumns(self, serieses: dict[str, pd.Series]):
        to_delete = set()
        to_assign: dict[str, pd.Series] = {}
        for k, v in serieses.items():
            if v is self.__delete:
                to_delete.add(k)
            else:
                to_assign[k] = v
        self._data_raw: pd.DataFrame = self._data_raw.assign(**to_assign).drop(
            to_delete, axis=1
        )
        nr, nc = self._data_raw.shape
        self.model().df = self._data_raw
        self.model().setShape(nr, nc)
        self._set_proxy(None)
        self.refreshTable()
        return None

    @assignColumns.server
    def assignColumns(self, serieses: dict[str, pd.Series]):
        columns = self._data_raw.columns
        old_param: dict[str, pd.Series] = {}
        for k in serieses.keys():
            if k in columns:
                old_param[k] = self._data_raw[k]
            else:
                old_param[k] = self.__delete
        return arguments(old_param)

    def createModel(self):
        model = DataFrameModel(self)
        self._qtable_view.setModel(model)
        return None

    def convertValue(self, c: int, value: Any) -> Any:
        """Convert value to the type of the table."""
        dtype = self._data_raw.dtypes[c]
        return get_converter(dtype)(value)



from __future__ import annotations
from enum import Enum, auto

import re
import ast
import sys
from typing import TYPE_CHECKING, cast
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt

from tabulous._qt._qt_const import MonospaceFontFamily
from tabulous._keymap import QtKeys
from tabulous.exceptions import UnreachableError
from tabulous.types import HeaderInfo, EvalInfo
from tabulous._range import RectRange, MultiRectRange
from tabulous._utils import get_config
from tabulous._selection_op import (
    find_last_dataframe_expr,
    construct,
    iter_extract,
)

if TYPE_CHECKING:
    from types import ModuleType
    from qtpy.QtCore import pyqtBoundSignal
    import pandas as pd
    from ._table_base import QMutableTable
    from ._enhanced_table import _QTableViewEnhanced
    from ._header_view import QDataFrameHeaderView

_LEFT_LIKE = frozenset(
    {"Left", "Ctrl+Left", "Shift+Left", "Ctrl+Shift+Left", "Home", "Shift+Home"}
)
_RIGHT_LIKE = frozenset(
    {"Right", "Ctrl+Right", "Shift+Right", "Ctrl+Shift+Right", "End", "Shift+End"}
)

MAC = sys.platform == "darwin"


class _QTableLineEdit(QtW.QLineEdit):
    """LineEdit widget with dtype checker and custom defocusing."""

    _VALID = QtGui.QColor(122, 122, 122, 200)
    _INVALID = QtGui.QColor(255, 0, 0, 200)

    def __init__(
        self,
        parent: QtCore.QObject | None = None,
        table: QMutableTable | None = None,
        pos: tuple[int, int] = (0, 0),
    ):
        super().__init__(parent)
        self._table = table
        self._pos = pos
        self._is_valid = True
        self._current_exception = ""
        self._attached_tooltip = ""
        self.textChanged.connect(self._on_text_changed)

    @property
    def current_exception(self) -> str:
        """The current exception that happened in the cell."""
        return self._current_exception

    @current_exception.setter
    def current_exception(self, value: str | Exception | None) -> None:
        if value is None:
            value = ""
        elif isinstance(value, Exception):
            # html colored text
            exc_type = type(value).__name__
            text = str(value).replace("<", "&lt;").replace(">", "&gt;")
            value = f"<b><font color='red'>{exc_type}</font></b>: {text}"
        self._current_exception = str(value)
        return self.setToolTip(self._current_exception)

    @classmethod
    def _is_eval_like(cls, text: str) -> bool:
        _eval_prefix = get_config().cell.eval_prefix
        _ref_prefix = get_config().cell.ref_prefix
        return text.startswith(_eval_prefix) or text.startswith(_ref_prefix)

    def parentTableView(self) -> _QTableViewEnhanced:
        """The parent table view."""
        return self.parent().parent()

    def _is_text_valid(self) -> bool:
        """True if text is valid for this cell."""
        raise NotImplementedError()

    def _on_text_changed(self, text: str) -> None:
        """Change text color to red if invalid."""
        self._is_valid = self._is_text_valid()
        return None

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        """Handle key press events."""
        keys = QtKeys(event)
        pos = self.cursorPosition()
        nchar = len(self.text())
        r, c = self._pos
        if keys.is_moving():
            if pos == 0 and keys == "Left" and c >= 0:
                self.parentTableView().setFocus()
                self._table._qtable_view._selection_model.move_to(r, c - 1)
                return
            elif (
                pos == nchar
                and keys == "Right"
                and c < self._table.model().columnCount() - 1
                and self.selectedText() == ""
            ):
                self.parentTableView().setFocus()
                self._table._qtable_view._selection_model.move_to(r, c + 1)
                return
            elif keys == "Up" and r >= 0:
                self.parentTableView().setFocus()
                self._table._qtable_view._selection_model.move_to(r - 1, c)
                return
            elif keys == "Down" and r < self._table.model().rowCount() - 1:
                self.parentTableView().setFocus()
                self._table._qtable_view._selection_model.move_to(r + 1, c)
                return

        elif keys == "F2":  # editing fails
            return None
        return super().keyPressEvent(event)

    def paintEvent(self, a0: QtGui.QPaintEvent) -> None:
        super().paintEvent(a0)
        rect = self.rect()
        rect.adjust(1, 1, -1, -1)
        p0 = rect.bottomLeft()
        p1 = rect.bottomRight()
        painter = QtGui.QPainter(self)
        if self._is_valid:
            painter.setPen(QtGui.QPen(self._VALID, 3))
        else:
            painter.setPen(QtGui.QPen(self._INVALID, 3))
        painter.drawLine(p0, p1)
        painter.end()
        return None

    def attachToolTip(self, text: str) -> None:
        """Attach a tooltip to the widget."""
        if text:
            point = self.mapToGlobal(self.rect().bottomLeft())
            point.setX(point.x() - 2)
            point.setY(point.y() - 10)
            QtW.QToolTip.showText(point, text, self)

        return None


class _QHeaderLineEdit(_QTableLineEdit):
    """Line edit used for editing header text."""

    ALIGNMENT: Qt.AlignmentFlag

    def _get_index(self) -> int:
        raise NotImplementedError()

    def _get_rect(self, index: int) -> QtCore.QRect:
        raise NotImplementedError()

    def _get_pandas_axis(self) -> pd.Index:
        raise NotImplementedError()

    def _get_signal(self) -> pyqtBoundSignal:
        raise NotImplementedError()

    def __init__(
        self,
        parent: QDataFrameHeaderView,
        table: QMutableTable | None = None,
        pos: tuple[int, int] = (0, 0),
    ):
        super().__init__(parent, table, pos)

        index = self._get_index()
        geometry = self._get_rect(index)
        geometry.adjust(2, 1, -2, -1)
        self.setGeometry(geometry)
        self.setAlignment(self.ALIGNMENT)
        self.setHidden(False)

        df_axis = self._get_pandas_axis()
        if index < df_axis.size:
            old_value = df_axis[index]
            text = str(old_value)
        else:
            # spreadsheet
            old_value = None
            text = ""

        @self.editingFinished.connect
        def _set_header_data():
            self.editingFinished.disconnect()
            value = self.text()
            err = None
            if not self._is_text_valid():
                err = ValueError(f"Duplicated name {value!r}")
            elif value != text and value:
                self._get_signal().emit(HeaderInfo(index, value, text))
            table = self._table._qtable_view
            table.setFocus()
            self.deleteLater()
            if err:
                raise err
            return None

        self.setText(text)
        self.selectAll()
        self.setFocus()

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        if event.key() == Qt.Key.Key_Escape:
            self.editingFinished.disconnect()
            self.hide()
            self.deleteLater()
            self._table._qtable_view.setFocus()
            return None
        return super().keyPressEvent(event)

    def _is_text_valid(self) -> bool:
        """True if text is valid for this cell."""
        text = self.text()
        pd_index = self._get_pandas_axis()
        idx = self._get_index()
        not_in = text not in pd_index
        if idx < pd_index.size:
            valid = text == pd_index[idx] or not_in
        else:
            valid = not_in

        if valid:
            self.current_exception = None
        else:
            self.current_exception = ValueError(f"Duplicated name {text!r}")
        return valid


class QVerticalHeaderLineEdit(_QHeaderLineEdit):
    """Line edit used for vertical editing header text."""

    ALIGNMENT = Qt.AlignmentFlag.AlignLeft

    def _get_index(self) -> int:
        return self._pos[0]

    def _get_rect(self, index: int) -> QtCore.QRect:
        header = self._table._qtable_view.verticalHeader()
        width = header.width()
        height = header.sectionSize(index)
        left = header.rect().left()
        top = header.sectionViewportPosition(index)
        return QtCore.QRect(left, top, width, height)

    def _get_pandas_axis(self) -> pd.Index:
        return self._table.model().df.index

    def _get_signal(self):
        return self._table.rowChangedSignal


class QHorizontalHeaderLineEdit(_QHeaderLineEdit):
    """Line edit used for horizontal editing header text."""

    ALIGNMENT = Qt.AlignmentFlag.AlignCenter

    def _get_index(self) -> int:
        return self._pos[1]

    def _get_rect(self, index: int) -> QtCore.QRect:
        header = self._table._qtable_view.horizontalHeader()
        width = header.sectionSize(index)
        height = header.height()
        left = header.sectionViewportPosition(index)
        top = header.rect().top()
        return QtCore.QRect(left, top, width, height)

    def _get_pandas_axis(self) -> pd.Index:
        return self._table.model().df.columns

    def _get_signal(self):
        return self._table.columnChangedSignal


class _EventFilter(QtCore.QObject):
    """An event filter for text completion by tab."""

    def eventFilter(self, o: QCellLiteralEdit, e: QtCore.QEvent):
        if e.type() == QtCore.QEvent.Type.KeyPress:
            e = cast(QtGui.QKeyEvent, e)
            if e.key() == Qt.Key.Key_Tab:
                o._on_tab_clicked()
                return True
        return False


class QCellLiteralEdit(_QTableLineEdit):
    """Line edit used for evaluating cell text."""

    class Mode(Enum):
        """Editing mode of the cell line edit."""

        TEXT = auto()
        EVAL = auto()

    def __init__(
        self,
        parent: QtCore.QObject | None = None,
        table: QMutableTable | None = None,
        pos: tuple[int, int] = (0, 0),  # the "visual" position
    ):
        super().__init__(parent, table, pos)
        if table is not None:
            self._source_pos = table._proxy.get_source_index(pos[0]), pos[1]
        else:
            self._source_pos = pos
        self._old_range: RectRange | None = None
        qtable = self.parentTableView()
        qtable._selection_model.moved.connect(self._on_selection_changed)
        qtable._focused_widget = self

        self._initial_rect = self.rect()
        self._self_focused = True
        self._completion_module: ModuleType | None = None

        self._event_filter = _EventFilter(self)
        self.installEventFilter(self._event_filter)

        self.mode = self.Mode.TEXT
        self.textEdited.connect(self._ensure_visible)

    @classmethod
    def from_table(
        cls,
        qtable: _QTableViewEnhanced,
        text: str,
    ) -> QCellLiteralEdit:
        """Create a new literal editor from a table."""
        parent = qtable.viewport()
        table = qtable.parentTable()
        rect = qtable.visualRect(
            qtable.model().index(*qtable._selection_model.current_index)
        )
        cr, cc = qtable._selection_model.current_index
        # cr = table._proxy.get_source_index(cr)
        line = cls(parent, table, (cr, cc))
        line.setMinimumSize(1, 1)
        geometry = line.geometry()
        geometry.setWidth(rect.width())
        geometry.setHeight(rect.height())
        geometry.moveCenter(rect.center())
        line.setGeometry(geometry)
        line.setText(text)

        if cls._is_eval_like(text):
            line.mode = cls.Mode.EVAL

        line.setFocus()
        line.selectAll()
        return line

    @classmethod
    def _parse_ref(cls, raw_text: str) -> tuple[str, bool]:
        """Convert texts if it starts with the evaluation prefixes."""
        _ref_prefix = get_config().cell.ref_prefix
        _eval_prefix = get_config().cell.eval_prefix
        if raw_text.startswith(_ref_prefix):
            text = raw_text.lstrip(_ref_prefix).strip()
            is_ref = True
        elif raw_text.startswith(_eval_prefix):
            text = raw_text.lstrip(_eval_prefix).strip()
            is_ref = False
        else:
            raise UnreachableError(raw_text)
        return text, is_ref

    @property
    def mode(self) -> Mode:
        """Edit mode."""
        return self._mode

    @mode.setter
    def mode(self, val: Mode) -> None:
        """Set edit mode."""
        self._mode = self.Mode(val)
        if self._mode is self.Mode.EVAL:
            font = QtGui.QFont(MonospaceFontFamily, self.font().pointSize())
            font.setBold(True)
            self.setFont(font)
        else:
            font_family = self.parentTableView().font().family()
            font = QtGui.QFont(font_family, self.font().pointSize())
            font.setBold(False)
            self.setFont(font)
        return None

    def _on_text_changed(self, text: str) -> None:
        """Change text color to red if invalid."""
        err_tip = ""
        if self._is_eval_like(text):
            self.mode = self.Mode.EVAL
        else:
            self.mode = self.Mode.TEXT

        self._manage_completion(text)
        self._reshape_widget(self.text())  # text may have changed!

        super()._on_text_changed(text)

        if self.mode is self.Mode.TEXT:
            self.attachToolTip("")
            return None

        # draw ranges
        _table = self._table
        ranges: list[tuple[slice, slice]] = []
        for op in iter_extract(text):
            ranges.append(op.as_iloc_slices(_table.model().df))

        if ranges:
            new_range = MultiRectRange.from_slices(ranges)
        else:
            new_range = None
        if self._old_range or new_range:
            _table._qtable_view._current_drawing_slot_ranges = new_range
            _table._qtable_view._update_all()
        self.attachToolTip(err_tip)
        return None

    def close(self) -> None:
        """Close this editor and deal with all the descruction."""
        qtable = self.parentTableView()
        qtable._selection_model.moved.disconnect(self._on_selection_changed)
        super().close()
        del qtable._focused_widget
        self.deleteLater()
        qtable.setFocus()
        # Need to emit moved signal again, otherwise table paint sometimes fails.
        qtable._selection_model.move_to(*qtable._selection_model.current_index)
        return None

    def eval_and_close(self) -> None:
        """Evaluate the text and close this editor."""
        if self._mode is self.Mode.TEXT:
            # setDataFrameValue itself considers proxy. Here we just need to
            # pass "_pos".
            self._table.setDataFrameValue(*self._pos, self.text())
            self.close()
        else:
            text, is_ref = self._parse_ref(self.text())
            info = EvalInfo(
                pos=self._pos,
                source_pos=self._source_pos,
                expr=text,
                is_ref=is_ref,
            )
            self._table.evaluatedSignal.emit(info)
        return None

    def _on_tab_clicked(self) -> None:
        """Tab-clicked event."""
        if self.mode is self.Mode.TEXT:
            # move right
            qtable = self.parentTableView()
            self.eval_and_close()
            qtable._selection_model.move(0, 1)
        elif self.mode is self.Mode.EVAL:
            # clear selection (= auto completion)
            nchar = len(self.text())
            self.setSelection(nchar, nchar)
        else:
            raise UnreachableError(self.mode)

    def _is_text_valid(self) -> bool:
        """Try to parse the text and return True if it is valid."""
        if self.mode is self.Mode.TEXT:
            _, c = self._pos
            try:
                convert_value = self._table._get_converter(c)
                convert_value(c, self.text())
            except Exception as e:
                self.current_exception = e
                return False
            else:
                self.current_exception = None
                return True
        elif self.mode is self.Mode.EVAL:
            raw_text = self.text()
            _ref_prefix = get_config().cell.ref_prefix
            _eval_prefix = get_config().cell.eval_prefix
            if raw_text.startswith(_eval_prefix):
                text = raw_text.lstrip(_eval_prefix).strip()
            elif raw_text.startswith(_ref_prefix):
                text = raw_text.lstrip(_ref_prefix).strip()
            else:
                return False

            if text == "":
                return True
            try:
                ast.parse(text)
            except Exception as e:
                self.current_exception = e
                return False
            else:
                self.current_exception = None
            return True
        else:
            raise UnreachableError(self.mode)

    def _ensure_visible(self, text: str) -> None:
        row, col = self._pos
        qtable = self.parentTableView()
        qtable.scrollTo(qtable.model().index(row, col), qtable.ScrollHint.EnsureVisible)

    def _on_selection_changed(self) -> None:
        """Update text based on the current selection."""
        qtable_view = self.parentTableView()
        if self.mode is self.Mode.TEXT:
            return self.eval_and_close()

        text = self.text()
        cursor_pos = self.cursorPosition()

        # prepare text
        if len(qtable_view._selection_model.ranges) != 1 or cursor_pos == 0:
            # If multiple cells are selected, don't update because selection range
            # is not well-defined. If cursor position is at the beginning, don't
            # update because it is before "=".
            return None

        qtable = cast("QMutableTable", qtable_view.parentTable())
        _df_filt = qtable_view.model().df
        _df_ori = qtable._data_raw
        rsl, csl = qtable_view._selection_model.ranges[-1]
        if not qtable._proxy.is_ordered and not _is_size_one(rsl):
            self.attachToolTip(
                "Table proxy is not ordered.\nCannot create cell "
                "references from table selection."
            )
            return None
        if rsl.stop is not None and rsl.stop > _df_filt.shape[0]:
            rsl = slice(rsl.start, _df_filt.shape[0])
        rsl = qtable._proxy.get_source_slice(rsl)
        table_range = RectRange.from_shape(_df_filt.shape)

        # out of border
        if not table_range.overlaps_with(RectRange(rsl, csl)):
            return None

        column_selected = len(qtable_view._selection_model._col_selection_indices) > 0
        selop = construct(
            rsl,
            csl,
            _df_ori,
            method="iloc",
            column_selected=column_selected,
        )

        if selop is None:  # out of bound
            return None

        # get string that represents the selection
        if selop.area(_df_ori) > 1:
            to_be_added = selop.fmt("df")
        else:
            to_be_added = selop.resolve_indices(_df_ori, (1, 1)).fmt_scalar("df")

        # add the representation to the text at the proper position
        if cursor_pos == 0:
            self.setText(to_be_added + text)
        elif text[cursor_pos - 1] != "]":
            self.setText(text[:cursor_pos] + to_be_added + text[cursor_pos:])
        else:
            idx = find_last_dataframe_expr(text[:cursor_pos])
            if idx >= 0:
                self.setText(text[:idx] + to_be_added + text[cursor_pos:])
            else:
                self.setText(text[:cursor_pos] + to_be_added + text[cursor_pos:])

        return None

    def _reshape_widget(self, text: str) -> None:
        """Resize to let all the characters visible."""
        fm = QtGui.QFontMetrics(self.font())
        width = min(fm.boundingRect(text).width() + 8, 300)
        return self.resize(max(width, self._initial_rect.width()), self.height())

    def _manage_completion(self, text: str) -> None:
        """Code completion."""
        if self.mode is self.Mode.TEXT:
            return None

        if text.endswith("."):
            mod_str = _find_last_module_name(text[:-1])
            if _viewer := self._table.parentViewer():
                mod = _viewer._namespace.get(mod_str, None)
                if mod is None:
                    self._completion_module = None
                    return None
                self._completion_module = mod
            return None

        elif self._completion_module is None:
            return None

        idx = text.rfind(".")
        if idx < 0:
            return None
        seed = text[idx + 1 :]
        for attr in self._completion_module.__dict__.keys():
            if attr.startswith(seed):
                current_text = self.text()
                current_pos = self.cursorPosition()
                new_text = current_text[: -len(seed)] + attr
                self.blockSignals(True)
                try:
                    self.setText(new_text)
                    self.setCursorPosition(current_pos)
                    self.setSelection(current_pos, len(new_text))
                finally:
                    self.blockSignals(False)
                break
        return None

    def _key_press_event(self, event: QtGui.QKeyEvent) -> None:
        keys = QtKeys(event)
        qtable = self.parentTableView()
        keys_str = str(keys)

        if keys.is_moving() or keys.is_moving_func():
            # cursor movements
            pos = self.cursorPosition()
            nchar = len(self.text())
            if not self._self_focused:
                # move in the parent table
                self._table._keymap.press_key(keys)
                self.setFocus()
                return True

            if keys_str in _LEFT_LIKE:
                # exit the editor if the cursor is at the beginning
                if pos == 0:
                    qtable._selection_model.move(0, -1)
                    self._self_focused = False
                    self.setFocus()
                    return True
                else:
                    self._self_focused = True
                    self.setFocus()
            elif keys_str in _RIGHT_LIKE:
                # exit the editor if the cursor is at the end
                if pos == nchar and self.selectedText() == "":
                    qtable._selection_model.move(0, 1)
                    self._self_focused = False
                    self.setFocus()
                    return True
                else:
                    self._self_focused = True
                    self.setFocus()

            else:
                self._self_focused = False

        elif keys == "Return":
            self.eval_and_close()
            return True
        elif keys == "Esc":
            qtable._selection_model.move_to(*self._pos)
            self.close()
            return True
        elif keys == "F2":  # editing fails
            return True
        elif keys == "F3":
            self.eval_and_close()
            editor = QCellLabelEdit.from_table(qtable)
            editor.show()
            return True

        if keys.is_typing() or keys_str in ("Backspace", "Delete"):
            if keys_str in ("Backspace", "Delete"):
                self._completion_module = None
            with qtable._selection_model.blocked():
                qtable._selection_model.move_to(*self._pos)
            self._self_focused = True
        self.setFocus()
        return False

    if MAC:

        def event(self, a0: QtCore.QEvent) -> bool:
            # NOTE: On MacOS, up/down causes text cursor to move to the beginning/end.
            _type = a0.type()
            if _type != QtCore.QEvent.Type.KeyPress:
                return super().event(a0)
            event = QtGui.QKeyEvent(a0)
            if not self._key_press_event(event):
                super().event(a0)
            self.setFocus()
            return True

    else:

        def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
            if not self._key_press_event(event):
                return QtW.QLineEdit.keyPressEvent(self, event)


_PATTERN_IDENTIFIERS = re.compile(r"[\w\d_]+")


def _find_last_module_name(s: str):
    return _PATTERN_IDENTIFIERS.findall(s)[-1]


class QCellLabelEdit(QtW.QLineEdit):
    """A QLineEdit that can be used to edit the label of a cell."""

    def __init__(
        self,
        parent: QtCore.QObject | None = None,
        table: QMutableTable | None = None,
        pos: tuple[int, int] = (0, 0),
    ):
        super().__init__(parent)
        self._table = table
        self._pos = pos
        self.editingFinished.connect(self._on_editing_finished)

    def _on_editing_finished(self):
        self.editingFinished.disconnect()
        self._hide_and_delete()
        text = self.text()
        if text == "":
            text = None
        self._table.setItemLabel(*self._pos, text)
        return None

    def _hide_and_delete(self):
        self.hide()
        self.deleteLater()
        self._table._qtable_view.setFocus()
        return None

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        keys = QtKeys(event)
        pos = self.cursorPosition()
        nchar = len(self.text())
        r, c = self._pos
        if keys == "Escape":
            self._hide_and_delete()
            return None
        if keys.is_moving():
            if pos == 0 and keys == "Left" and c >= 0:
                self._on_editing_finished()
                self._table._qtable_view._selection_model.move_to(r, c - 1)
                return
            elif (
                pos == nchar
                and keys == "Right"
                and c < self._table.model().columnCount() - 1
                and self.selectedText() == ""
            ):
                self._on_editing_finished()
                self._table._qtable_view._selection_model.move_to(r, c + 1)
                return
            elif keys == "Up" and r >= 0:
                self._on_editing_finished()
                self._table._qtable_view._selection_model.move_to(r - 1, c)
                return
            elif keys == "Down" and r < self._table.model().rowCount() - 1:
                self._on_editing_finished()
                self._table._qtable_view._selection_model.move_to(r + 1, c)
                return

        elif keys == "F2":
            return None
        elif keys == "F3":
            return None  # do nothing
        return super().keyPressEvent(event)

    @classmethod
    def from_table(
        cls,
        qtable: _QTableViewEnhanced,
    ) -> QCellLiteralEdit:
        """Create a new literal editor from a table."""
        parent = qtable.viewport()
        table = qtable.parentTable()
        index = qtable._selection_model.current_index
        rect = qtable.visualRect(qtable.model().index(*index))
        line = cls(parent, table, index)
        line.setMinimumSize(1, 1)
        geometry = line.geometry()
        geometry.setWidth(rect.width())
        geometry.setHeight(rect.height())
        geometry.moveCenter(rect.center())
        line.setGeometry(geometry)

        text = table.itemLabel(*index)
        line.setText(text)

        line.setFocus()
        line.selectAll()

        font = qtable.font()
        font.setBold(True)
        line.setFont(font)
        return line


def _is_size_one(sl: slice) -> bool:
    if sl.start is None or sl.stop is None:
        return False
    return sl.start == sl.stop - 1



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtCore import Qt, Signal


class QTitleBar(QtW.QWidget):
    """A custom title bar for a QSplitterDockWidget"""

    closeSignal = Signal()

    def __init__(self, title: str = "", parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        _layout = QtW.QHBoxLayout()
        _layout.setContentsMargins(4, 0, 4, 0)
        _layout.setSpacing(0)

        self._title_label = QtW.QLabel()
        self._title_label.setContentsMargins(0, 0, 0, 0)

        _frame = QtW.QFrame()
        _frame.setFrameShape(QtW.QFrame.Shape.HLine)
        _frame.setFrameShadow(QtW.QFrame.Shadow.Sunken)
        _frame.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Fixed
        )
        self._close_button = QtW.QToolButton()
        self._close_button.setText("✕")
        self._close_button.setToolTip("Close the widget.")
        self._close_button.setFixedSize(QtCore.QSize(16, 16))
        self._close_button.setCursor(Qt.CursorShape.ArrowCursor)

        _layout.addWidget(self._title_label)
        _layout.addWidget(_frame)
        _layout.addWidget(self._close_button)
        _layout.setAlignment(self._close_button, Qt.AlignmentFlag.AlignRight)
        self.setLayout(_layout)

        self._close_button.clicked.connect(self.closeSignal.emit)

        self.setTitle(title)
        self.setFixedHeight(18)

    def title(self) -> str:
        """The title text."""
        return self._title_label.text()

    def setTitle(self, text: str):
        """Set the title text."""
        if text == "":
            self._title_label.setVisible(False)
        else:
            self._title_label.setVisible(True)
            self._title_label.setText(f"  {text}  ")
